#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Feb 22, 2010 by: rch

from traits.api import \
    HasTraits, Instance, Int, Property, Array, cached_property, Instance

from numpy import \
     zeros, dot, hstack, vstack, identity, cross, transpose, tensordot, outer

from scipy.linalg import \
     inv, norm

from ibvpy.fets.fets_eval import FETSEval
from ibvpy.fets.fets3D.fets3D8h import FETS3D8H

# import the 3D elastic matrix in order  
from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import \
    MATS3DElastic

from ibvpy.tmodel.mats2D.mats2D_tensor import \
    map2d_eps_mtx_to_eng, map2d_sig_eng_to_mtx

from ibvpy.tmodel.mats3D.mats3D_tensor import \
    map3d_eps_eng_to_mtx, map3d_sig_eng_to_mtx, map3d_sig_mtx_to_eng, map3d_eps_mtx_to_eng, map3d_tns2_to_tns4, map3d_tns4_to_tns2, map3d_eps_mtx_to_eng

#from i_fets_eval import IFETSEval

#-----------------------------------------------------------------------------------
# FETS2D58H -  degenerated 8 nodes isoparametric volume element (2D, linear, Lagrange family)    
#              using a 2D material model
#-----------------------------------------------------------------------------------


class FETS2D5( HasTraits ):

    # @todo: the calculation of the direction cosines 
    #        should be cached (similar to J_mtx and B_mtx)


    elastic_D_mtx3D = Property( depends_on = 'mats_eval' )
    @cached_property
    def _get_elastic_D_mtx3D( self ):
        elastic_mats_eval = MATS3DElastic( E = self.mats_eval.E, nu = self.mats_eval.nu )
        zero_3D = zeros( ( 6, ) )
        sig_eng3D, D_mtx3D = elastic_mats_eval.get_corr_pred( None, zero_3D, zero_3D, 0, 0 )
        return D_mtx3D

    def get_dircos_mtx( self, sctx ):
        '''returns the direction cosine at the integration point.
        The three column vectors v1, v2, v3 of the direction cosine
        matrix build an orthonormal system and can be used to rotate
        the 3D strains into the local coordinate system. 
        The vector v3 stands perpendicular to the plane defined by the 
        curvilinear coordinates r and s. Note that for the orientation
        of the local in-plane coordinates different choices exist.'''

        # get the Jacobian matrix
        r_pnt = sctx.r_pnt
        X_mtx = sctx.X
        J_mtx = self.get_J_mtx( r_pnt, X_mtx )

        # the first two rows of the Jacobi matrix are tangential
        # to the mid-surface at the integration point.  
        # (cf. Zienkiewicz Vol.2, Eq.(15.19))
        j1 = J_mtx[0, :]
        j2 = J_mtx[1, :]

        # v3 stands perdendicular to the surface
        v3 = cross( j1, j2 )

        # One possible choice for the direction of the local x-axis 
        # is to define it such that it coincides with the orientation
        # of the first curvilinear coordinate r (cf. FE-Skript)
        v1 = j1

        # The local y-axis is obtained as the cross product of the 
        # local x and z direction
        v2 = cross( v3, v1 )

        # normalize the vectors to obtain an ortho-normal basis 
        v1_ = v1 / norm( v1 )
        v2_ = v2 / norm( v2 )
        v3_ = v3 / norm( v3 )

        return transpose( vstack( ( v1_, v2_, v3_ ) ) )


    def get_Trans_strain_to_loc_mtx( self, dircos_mtx ):
        '''Transformation matrix in 3d shape = (6,6).
        multiply the strain vector (engineering notation) with 
        the returned matrix in order to obtain the strain vector
        (engineering notation) in the local coordinate system.
        The ordering of the strain vector in engineering notation
        corresponds to the ordering 
        [eps11,eps22,eps33,eps23,eps13,eps12]. 
        The parameter 'dircos' is the matrix of the direction
        cosine describing the relation between the global and 
        local coordinate system.
        (cf. Zienkiewicz Eq.(6.24))
        '''
        N = dircos_mtx
        T = zeros( ( 6, 6 ) )
        T[0, 0] = N[0, 0] * N[0, 0]
        T[0, 1] = N[1, 0] * N[1, 0]
        T[0, 2] = N[2, 0] * N[2, 0]
        T[0, 3] = N[1, 0] * N[2, 0]
        T[0, 4] = N[0, 0] * N[2, 0]
        T[0, 5] = N[0, 0] * N[1, 0]
        T[1, 0] = N[0, 1] * N[0, 1]
        T[1, 1] = N[1, 1] * N[1, 1]
        T[1, 2] = N[2, 1] * N[2, 1]
        T[1, 3] = N[1, 1] * N[2, 1]
        T[1, 4] = N[0, 1] * N[2, 1]
        T[1, 5] = N[0, 1] * N[1, 1]
        T[2, 0] = N[0, 2] * N[0, 2]
        T[2, 1] = N[1, 2] * N[1, 2]
        T[2, 2] = N[2, 2] * N[2, 2]
        T[2, 3] = N[1, 2] * N[2, 2]
        T[2, 4] = N[0, 2] * N[2, 2]
        T[2, 5] = N[0, 2] * N[1, 2]
        T[3, 0] = 2.0 * N[0, 2] * N[0, 1]
        T[3, 1] = 2.0 * N[1, 2] * N[1, 1]
        T[3, 2] = 2.0 * N[2, 2] * N[2, 1]
        T[3, 3] = N[1, 2] * N[2, 1] + N[2, 2] * N[1, 1]
        T[3, 4] = N[2, 2] * N[0, 1] + N[0, 2] * N[2, 1]
        T[3, 5] = N[0, 2] * N[1, 1] + N[1, 2] * N[0, 1]
        T[4, 0] = 2.0 * N[0, 2] * N[0, 0]
        T[4, 1] = 2.0 * N[1, 2] * N[1, 0]
        T[4, 2] = 2.0 * N[2, 2] * N[2, 0]
        T[4, 3] = N[1, 2] * N[2, 0] + N[2, 2] * N[1, 0]
        T[4, 4] = N[2, 2] * N[0, 0] + N[0, 2] * N[2, 0]
        T[4, 5] = N[0, 2] * N[1, 0] + N[1, 2] * N[0, 0]
        T[5, 0] = 2.0 * N[0, 1] * N[0, 0]
        T[5, 1] = 2.0 * N[1, 1] * N[1, 0]
        T[5, 2] = 2.0 * N[2, 1] * N[2, 0]
        T[5, 3] = N[1, 1] * N[2, 0] + N[2, 1] * N[1, 0]
        T[5, 4] = N[2, 1] * N[0, 0] + N[0, 1] * N[2, 0]
        T[5, 5] = N[0, 1] * N[1, 0] + N[1, 1] * N[0, 0]
        Trans_strain_to_loc_mtx = T
        return Trans_strain_to_loc_mtx


    def get_Trans_stress_from_loc_mtx( self, dircos_mtx ):
        '''Transformation matrix in 3d shape = (6,6)  
        multiply the local stress vector (engineering notation) with 
        the returned matrix in order to obtain the stress vector
        (engineering notation) in the global coordinate system.
        The ordering of the stress vector in engineering notation
        corresponds to the ordering [sig11,sig22,sig33,sig23,sig13,sig12]. 
        The parameter 'dircos' is the matrix of the direction
        cosine describing the relation between the global and 
        local coordinate system.
        (cf. Zienkiewicz Eq.(6.23))
        '''
        N = dircos_mtx
        T = zeros( ( 6, 6 ) )
        T[0, 0] = N[0, 0] * N[0, 0]
        T[0, 1] = N[0, 1] * N[0, 1]
        T[0, 2] = N[0, 2] * N[0, 2]
        T[0, 3] = 2.0 * N[0, 1] * N[0, 2]
        T[0, 4] = 2.0 * N[0, 0] * N[0, 2]
        T[0, 5] = 2.0 * N[0, 0] * N[0, 1]
        T[1, 0] = N[1, 0] * N[1, 0]
        T[1, 1] = N[1, 1] * N[1, 1]
        T[1, 2] = N[1, 2] * N[1, 2]
        T[1, 3] = 2.0 * N[1, 1] * N[1, 2]
        T[1, 4] = 2.0 * N[1, 0] * N[1, 2]
        T[1, 5] = 2.0 * N[1, 0] * N[1, 1]
        T[2, 0] = N[2, 0] * N[2, 0]
        T[2, 1] = N[2, 1] * N[2, 1]
        T[2, 2] = N[2, 2] * N[2, 2]
        T[2, 3] = 2.0 * N[2, 1] * N[2, 2]
        T[2, 4] = 2.0 * N[2, 0] * N[2, 2]
        T[2, 5] = 2.0 * N[2, 0] * N[2, 1]
        T[3, 0] = N[2, 0] * N[1, 0]
        T[3, 1] = N[2, 1] * N[1, 1]
        T[3, 2] = N[2, 2] * N[1, 2]
        T[3, 3] = N[2, 1] * N[1, 2] + N[2, 2] * N[1, 1]
        T[3, 4] = N[2, 2] * N[1, 0] + N[2, 0] * N[1, 2]
        T[3, 5] = N[2, 0] * N[1, 1] + N[2, 1] * N[1, 0]
        T[4, 0] = N[2, 0] * N[0, 0]
        T[4, 1] = N[2, 1] * N[0, 1]
        T[4, 2] = N[2, 2] * N[0, 2]
        T[4, 3] = N[2, 1] * N[0, 2] + N[2, 2] * N[0, 1]
        T[4, 4] = N[2, 2] * N[0, 0] + N[2, 0] * N[0, 2]
        T[4, 5] = N[2, 0] * N[0, 1] + N[2, 1] * N[0, 0]
        T[5, 0] = N[1, 0] * N[0, 0]
        T[5, 1] = N[1, 1] * N[0, 1]
        T[5, 2] = N[1, 2] * N[0, 2]
        T[5, 3] = N[1, 1] * N[0, 2] + N[1, 2] * N[0, 1]
        T[5, 4] = N[1, 2] * N[0, 0] + N[1, 0] * N[0, 2]
        T[5, 5] = N[1, 0] * N[0, 1] + N[1, 1] * N[0, 0]
        Trans_stress_from_loc_mtx = T
        return Trans_stress_from_loc_mtx

    def get_eps_eng( self, sctx, u ):

        eps_eng = super( FETS2D58H, self ).get_eps_eng( sctx, u )
        # get the direction cosine at the ip:
        dircos_mtx = self.get_dircos_mtx( sctx )
        dircos_mtx_T = transpose( dircos_mtx )

        # switch from engineering notation to matrix notation:
        eps_mtx3D = map3d_eps_eng_to_mtx( eps_eng )

        # transform strain from global to local coordinates (i.e. map strain into the 2D plane) 
        eps_mtx3D_ = dot( dot( dircos_mtx_T, eps_mtx3D ), dircos_mtx )

        # reduce 3D mtx to 2D mtx (neglecting the strains in local z_ direction!)
        eps_mtx2D_ = eps_mtx3D_[:2, :2]

        # switch from matrix notation to engineering notation for the strains:
        eps_eng2D_ = map2d_eps_mtx_to_eng( eps_mtx2D_ )
        return eps_eng2D_

    def get_mtrl_corr_pred( self, sctx, eps_eng, d_eps_eng, tn, tn1, eps_avg = None ):
        '''Overload the 3D material-corrector-predictor and 
        use the 2D material model instead. This requires to
        map the 3D strains into a 2D plane. This is performed
        employing the direction cosines at each integration
        point (ip). The direction cosines establish the link
        between the global and local coordinate system (the
        reference to the local coordinate system is indicated 
        by the underline character at the end of a variable
        name). 
        After employing the 2D material model the returned 
        stresses and stiffness are augmented with the z_-components
        and transformed into global coordinates. It is assumed
        that in global z-direction the material behaves linear-elastic.
        @todo: check for consistency: 
        
        var A: degenerated volume (assume plane strain eps_z_ == 0)
               for the evaluation of the 2D material model and the 
               damage in the in-plane elasticity components. 
               Then correct the stresses by adding the eps_z_ != 0 
               part. For the linear elastic regime this ends up to
               the standard 3D formulation.
               @todo: check for consistency:
               the damage functions have been obtained based on the 
               assumtion for plane stress (which corresponds to the 
               situation in the experiment). Can the 2D case be 
               reproduced with the choosen implementation? 
        
        '''

        # get the direction cosine at the ip:
        dircos_mtx = self.get_dircos_mtx( sctx )
        dircos_mtx_T = transpose( dircos_mtx )

        # get the Transformation matrices for switching between the 
        # local and global coordinate system in 3D case:
        Trans_stress_from_loc_mtx = self.get_Trans_stress_from_loc_mtx( dircos_mtx )
        Trans_strain_to_loc_mtx = self.get_Trans_strain_to_loc_mtx( dircos_mtx )

        # switch from engineering notation to matrix notation:
        eps_mtx3D = map3d_eps_eng_to_mtx( eps_eng )
        d_eps_mtx3D = map3d_eps_eng_to_mtx( d_eps_eng )

        # transform strain from global to local coordinates (i.e. map strain into the 2D plane) 
        eps_mtx3D_ = dot( dot( dircos_mtx_T, eps_mtx3D ), dircos_mtx )
        d_eps_mtx3D_ = dot( dot( dircos_mtx_T, d_eps_mtx3D ), dircos_mtx )

        # reduce 3D mtx to 2D mtx (neglecting the strains in local z_ direction!)
        eps_mtx2D_ = eps_mtx3D_[:2, :2]
        d_eps_mtx2D_ = d_eps_mtx3D_[:2, :2]
#        print 'eps_mtx2D_: ', eps_mtx2D_

        # switch from matrix notation to engineering notation for the strains:
        eps_eng2D_ = map2d_eps_mtx_to_eng( eps_mtx2D_ )
        d_eps_eng2D_ = map2d_eps_mtx_to_eng( d_eps_mtx2D_ )
#        print 'eps_eng2D_: ', eps_eng2D_

        ###
#       # @todo: verify implementation with alternative variant using
#                    Transformation matrix and engineering notation:
#        eps_eng3D_   = dot( Trans_strain_to_loc_mtx,  eps_eng  )
#        d_eps_eng3D_ = dot( Trans_strain_to_loc_mtx, d_eps_eng )
        ###

        # evaluate the 2D mtrl-model
        sig_eng2D_, D_mtx2D_ = self.mats_eval.get_corr_pred( sctx, eps_eng2D_, d_eps_eng2D_, tn, tn1 )

        #------------------------------------------------------------------------
        # get predictor (D_mtx)        
        #------------------------------------------------------------------------

        # Take the initial elastic isotropic matrix as an initial value
        # the submatrix returned by the 2D damage model shall be written
        # into the slice [0:2,0:2] and [5,5]
        #
        D_mtx3D_ = self.elastic_D_mtx3D
        D_mtx3D_[:2, :2] = D_mtx2D_[:2, :2]
        D_mtx3D_[5, 5] = D_mtx2D_[2, 2]
#        print 'D_mtx3D_', D_mtx3D_

        # transform D_mtx3D from local to global coordinates 
        D_mtx3D = dot( Trans_stress_from_loc_mtx, dot( D_mtx3D_, Trans_strain_to_loc_mtx ) )
#        print 'D_mtx3D', D_mtx3D

        # Alternative: use numpy functionality and 4th-order tensors
        # @todo: varify the numpy-alternative with the existing implementation
#        dircos_tns4 = outer( dircos_mtx, dircos_mtx ).reshape(3,3,3,3) 
#        print 'dircos_tns4', dircos_tns4.shape
#        D_tns4_3D_ = map3d_tns2_to_tns4 ( D_mtx3D_ )
#        D_tns4_3D  = tensordot( dircos_tns4, tensordot( D_tns4_3D_, dircos_tns4, [[2,3],[2,3]] ), [[2,3],[0,1]] )
#        D_mtx3D    = map3d_tns4_to_tns2( D_tns4_3D )        


        #------------------------------------------------------------------------
        # get corrector (sig_eng)        
        #------------------------------------------------------------------------

        # augment 2D matrices to the 3D-case: 
        # the material behaves linear-elastic in out-of plane directions (xz, yz);

        # linear-elastic response (used for eps_zz_, eps_yz_, eps_xz_):
        eps_eng3D_ = map3d_eps_mtx_to_eng( eps_mtx3D_ )
        sig_eng3D_ = dot( D_mtx3D_, eps_eng3D_ )

        # replace the values for xx_, yy_, and xy_ based on the evaluation of mtrl2d:
        # and correct/superpose these values by the elastic response caused by eps_zz_: 
        sig_eng3D_[0] = sig_eng2D_[0] + D_mtx3D_[0, 2] * eps_eng3D_[2]
        sig_eng3D_[1] = sig_eng2D_[1] + D_mtx3D_[1, 2] * eps_eng3D_[2]
        sig_eng3D_[5] = sig_eng2D_[2]

        # transform sig_mtx3D from local to global coordinates
        sig_mtx3D_ = map3d_sig_eng_to_mtx( sig_eng3D_ )
        sig_mtx3D = dot( dot( dircos_mtx, sig_mtx3D_ ), dircos_mtx_T )

        # switch from matrix notations to engineering notations for the stresses:
        sig_eng3D = map3d_sig_mtx_to_eng( sig_mtx3D )

        return sig_eng3D, D_mtx3D

#----------------------- example --------------------

if __name__ == '__main__':
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
        TLine, BCDofGroup, IBVPSolve as IS, DOTSEval

#    from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import MATS3DElastic
    from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic

#    fets_eval = FETS3D8H(mats_eval = MATS3DElastic(nu = 0.25))
    fets_eval = FETS2D58H( mats_eval = MATS2DElastic( nu = 0.25, stress_state = "plane_strain" ) )

    # set vtk points for shrinked mode visualization:             
    fets_eval.vtk_r = fets_eval.vtk_r * 1.0

    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid( coord_max = ( 3., 3., 3. ),
                     shape = ( 1, 1, 1 ),
                     fets_eval = fets_eval )

    ts = TS( 
            sdomain = domain,
             bcond_list = [BCDofGroup( var = 'u', value = 0., dims = [0],
                                  get_dof_method = domain.get_left_dofs ),
                        BCDofGroup( var = 'u', value = 0., dims = [1, 2],
                                  get_dof_method = domain.get_bottom_left_dofs ),
                        BCDofGroup( var = 'u', value = 0.002, dims = [0],
                                  get_dof_method = domain.get_right_dofs ) ],
             rtrace_list = [
#                        RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
#                                  var_y = 'F_int', idx_y = right_dof,
#                                  var_x = 'U_k', idx_x = right_dof,
#                                  record_on = 'update'),
#                        RTraceDomainListField(name = 'Deformation' ,
#                                       var = 'eps', idx = 0,
#                                       record_on = 'update'),
                         RTraceDomainListField( name = 'Displacement' ,
                                        var = 'u', idx = 0, warp = True ),
#                         RTraceDomainListField(name = 'Stress' ,
#                                        var = 'sig', idx = 0,
#                                        record_on = 'update'),
#                        RTraceDomainListField(name = 'N0' ,
#                                       var = 'N_mtx', idx = 0,
#                                       record_on = 'update')
                        ]
            )

    # Add the time-loop control
    tloop = TLoop( tstepper = ts, KMAX = 4, RESETMAX = 0, tolerance = 1e-3,
         tline = TLine( min = 0.0, step = 1.0, max = 1.0 ) )

    print('u', tloop.eval())

    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp( ibv_resource = tloop )
    app.main()



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jun 30, 2009 by: rchx

if __name__ == '__main__':
    def example_1d():
        from ibvpy.api import FEDomain, FERefinementGrid, FEGrid, TStepBC as TS, \
            BCDofGroup, RTraceDomainListField, TLoop, TLine
        from ibvpy.mesh.xfe_subdomain import XFESubDomain
        from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic
        from ibvpy.fets.fets1D.fets1D2l import FETS1D2L
        from ibvpy.fets.fets1D.fets1D2l3u import FETS1D2L3U
        from ibvpy.fets.fets_ls.fets_crack import FETSCrack
        fets_eval = FETS1D2L( mats_eval = MATS1DElastic( E = 1. ) ) #, A=1.))
        #xfets_eval = fets_eval # use the same element for the enrichment
        xfets_eval = FETSCrack( parent_fets = fets_eval )
        # Discretization

        fe_domain = FEDomain()
        fe_level1 = FERefinementGrid( domain = fe_domain, fets_eval = fets_eval )
        fe_grid1 = FEGrid( coord_max = ( 4., 0., 0. ),
                           shape = ( 4, ),
                           fets_eval = fets_eval,
                           level = fe_level1 )


        enr = True
        if enr:
            fe_xdomain = XFESubDomain( domain = fe_domain,
                                       fets_eval = xfets_eval,
                                       fe_grid_slice = fe_grid1[  '(X - 2) **2 - 0.5 ' ] )
            fe_xdomain.deactivate_sliced_elems()

        print('elem_dof_map', fe_xdomain.elem_dof_map)

        fe_domain = FEDomain()
        fe_level1 = FERefinementGrid( domain = fe_domain, fets_eval = fets_eval )
        fe_grid1 = FEGrid( coord_max = ( 4 * 3.14, 0., 0. ),
                           shape = ( 8, ),
                           fets_eval = fets_eval,
                           level = fe_level1 )


        enr = True
        if enr:
            fe_xdomain = XFESubDomain( domain = fe_domain,
                                       fets_eval = xfets_eval,
                                       fe_grid_slice = fe_grid1[  'cos(X) - 0.5' ] )
            fe_xdomain.deactivate_sliced_elems()

        print('elem_dof_map2', fe_xdomain.elem_dof_map)


if __name__ == '__main__':
    example_1d()



'''
Created on Sep 3, 2009

@author: RChudoba, ABaktheer
'''

from ibvpy.tmodel.mats2D.mats2D_tensor import \
    map2d_sig_eng_to_mtx, map2d_eps_mtx_to_eng, map2d_sig_mtx_to_eng, \
    map2d_tns4_to_tns2, compliance_mapping2d
from ibvpy.tmodel.mats2D.mats2D_tensor import map2d_eps_eng_to_mtx
from traits.api import \
    Array, Property, cached_property, Callable, Constant
from bmcs_utils.api import Enum, Float, View, Item

from ibvpy.tmodel.matsXD.vmatsXD_eval import MATSXDEval
import numpy as np


class MATS2DEval(MATSXDEval):

    n_dims = Constant(2)

    ipw_view = View(
        Item('E'),
        Item('nu'),
        Item('eps_max'),
        Item('G_f', latex=r'G_\mathrm{f}^{\mathrm{estimate}}', readonly=True),
    )

    G_f = Property(Float, depends_on='state_changed')
    @cached_property
    def _get_G_f(self):
        eps_max = self.eps_max
        n_eps = 1000
        eps11_range = np.linspace(1e-9,eps_max,n_eps)
        eps_range = np.zeros((len(eps11_range), 2, 2))
        eps_range[:, 1, 1] = eps11_range
        state_vars = { var : np.zeros( (len(eps11_range),) + shape )
            for var, shape in self.state_var_shapes.items()
        }
        sig_range, _ = self.get_corr_pred(eps_range, 1, **state_vars)
        sig11_range = sig_range[:,1,1]
        return np.trapz(sig11_range, eps11_range)

    stress_state = Enum(
        options=['plane_stress', 'plane_strain'], MAT=True
    )

    D_ab = Property(Array, depends_on='MAT')
    '''Elasticity matrix (shape: (3,3))
    '''
    @cached_property
    def _get_D_ab(self):
        if self.stress_state == 'plane_stress':
            return self._get_D_ab_plane_stress()
        elif self.stress_state == 'plane_strain':
            return self._get_D_ab_plane_strain()

    def _get_D_ab_plane_stress(self):
        '''
        Elastic Matrix - Plane Stress
        '''
        E = self.E
        nu = self.nu
        D_stress = np.zeros([3, 3])
        D_stress[0, 0] = E / (1.0 - nu * nu)
        D_stress[0, 1] = E / (1.0 - nu * nu) * nu
        D_stress[1, 0] = E / (1.0 - nu * nu) * nu
        D_stress[1, 1] = E / (1.0 - nu * nu)
        D_stress[2, 2] = E / (1.0 - nu * nu) * (1.0 / 2.0 - nu / 2.0)
        return D_stress

    def _get_D_ab_plane_strain(self):
        '''
        Elastic Matrix - Plane Strain
        '''
        E = self.E
        nu = self.nu
        D_strain = np.zeros([3, 3])
        D_strain[0, 0] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[0, 1] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1, 0] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1, 1] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[2, 2] = E * (1.0 - nu) / (1.0 + nu) / (2.0 - 2.0 * nu)
        return D_strain

    map2d_ijkl2a = Array(np.int_,
                         value=[[[[0, 0],
                                  [0, 0]],
                                 [[2, 2],
                                  [2, 2]]],
                                [[[2, 2],
                                  [2, 2]],
                                 [[1, 1],
                                  [1, 1]]]])
    map2d_ijkl2b = Array(np.int_,
                         value=[[[[0, 2],
                                  [2, 1]],
                                 [[0, 2],
                                  [2, 1]]],
                                [[[0, 2],
                                  [2, 1]],
                                 [[0, 2],
                                  [2, 1]]]])

    D_abcd = Property(Array, depends_on='MAT')

    @cached_property
    def _get_D_abcd(self):
        return self.D_ab[self.map2d_ijkl2a, self.map2d_ijkl2b]


class NotUsed:
    # dimension-dependent mappings
    #
    map_tns4_to_tns2 = Callable(map2d_tns4_to_tns2, transient=True)
    map_eps_eng_to_mtx = Callable(map2d_eps_eng_to_mtx, transient=True)
    map_sig_eng_to_mtx = Callable(map2d_sig_eng_to_mtx, transient=True)
    compliance_mapping = Callable(compliance_mapping2d, transient=True)
    map_sig_mtx_to_eng = Callable(map2d_sig_mtx_to_eng, transient=True)
    map_eps_mtx_to_eng = Callable(map2d_eps_mtx_to_eng, transient=True)

    def _get_explorer_config(self):
        '''Get the specific configuration of this material model in the explorer
        '''
        c = super(MATS2DEval, self)._get_explorer_config()

        from ibvpy.api import TLine
        from ibvpy.tmodel.mats2D.mats2D_explorer_bcond import BCDofProportional

        # overload the default configuration
        c['bcond_list'] = [
            BCDofProportional(max_strain=0.00016, alpha_rad=np.pi / 8.0)]
        c['tline'] = TLine(step=0.05, max=1)
        return c




'''
Created on 05.12.2016

@author: abaktheer
'''

from os.path import join

from ibvpy.tmodel import MATSEval
from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, LiDamageFn, JirasekDamageFn, AbaqusDamageFn, \
    MultilinearDamageFn, \
    FRPDamageFn
from ibvpy.mathkit.mfn.mfn_line.mfn_line import MFnLineArray
from traits.api import  \
    Tuple, List, on_trait_change, \
    Instance, Trait, Bool, Str, Button, Property, cached_property

import bmcs_utils.api as bu
import numpy as np

import ipyregulartable as rt

class MATSBondSlipTriLinear(MATSEval):
    """Multilinear bond-slip law
    """
    name = "tri-linear bond law"

    E_m = bu.Float(28000.0, tooltip='Stiffness of the matrix [MPa]',
                MAT=True, unit='MPa', symbol=r'E_\mathrm{m}',
                desc='E-modulus of the matrix',
                auto_set=True, enter_set=True)

    E_f = bu.Float(170000.0, tooltip='Stiffness of the fiber [MPa]',
                MAT=True, unit='MPa', symbol=r'E_\mathrm{f}',
                desc='E-modulus of the reinforcement',
                auto_set=False, enter_set=True)

    tau_1 = bu.Float(10.0, tooltip='Shear strength [MPa]',
                MAT=True, unit='MPa', symbol=r'\tau_1',
                desc='shear strength',
                auto_set=False, enter_set=True)

    tau_2 = bu.Float(1.0, tooltip='Shear at plateau [MPa]',
                MAT=True, unit='MPa', symbol=r'\tau_2',
                desc='shear plateau',
                auto_set=False, enter_set=True)

    s_1 = bu.Float(0.1, tooltip='Slip at peak [mm]',
                MAT=True, unit='mm', symbol='s_1',
                desc='slip at peak',
                auto_set=False, enter_set=True)

    s_2 = bu.Float(0.5, tooltip='Slip at plateau [mm]',
                MAT=True, unit='mm', symbol='s_2',
                desc='slip at plateau',
                auto_set=False, enter_set=True)


    s_tau_table = Property(depends_on='state_changed')
    @cached_property
    def _get_s_tau_table(self):
        s_data = [0, self.s_1, self.s_2, self.s_2 * 3]
        tau_data = [0, self.tau_1, self.tau_2, self.tau_2]
        if len(s_data) != len(tau_data):
            raise ValueError('s array and tau array must have the same size')
        return s_data, tau_data

    #=========================================================================
    # Configurational parameters
    #=========================================================================
    U_var_shape = (1,)
    '''Shape of the primary variable required by the TStepState.
    '''

    state_var_shapes = {}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    node_name = 'tri-linear linear bond'

    def get_corr_pred(self, s, t_n1):

        n_e, n_ip, _ = s.shape
        D = np.zeros((n_e, n_ip, 3, 3))
        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f

        tau = np.einsum('...st,...t->...s', D, s)
        s = s[..., 1]
        shape = s.shape
        signs = np.sign(s.flatten())
        s_pos = np.fabs(s.flatten())
        bs_law = self.bs_law
        tau[..., 1] = (signs * bs_law(s_pos)).reshape(*shape)
        D_tau = self.bs_law.diff(s_pos).reshape(*shape)
        D[..., 1, 1] = D_tau

        return tau, D

    bs_law = Property(depends_on='state_changed')
    @cached_property
    def _get_bs_law(self):
        s_data, tau_data = self.s_tau_table
        return MFnLineArray(
            xdata=s_data,
            ydata=tau_data,
            plot_diff=False
        )

    def plot(self, ax, **kw):
        s_data, tau_data = self.s_tau_table
        ax.plot(s_data, tau_data, **kw)
        ax.fill_between(s_data, tau_data, alpha=0.1, **kw)
        ax.set_xlabel(r'$\tau$ [MPa]')
        ax.set_xlabel(r'$s$ [mm]')

    ipw_view = bu.View(
        bu.Item('E_m'),
        bu.Item('E_f'),
        bu.Item('tau_1'),
        bu.Item('s_1'),
        bu.Item('tau_2'),
        bu.Item('s_2'),
    )

    def update_plot(self, axes):
        self.plot(axes)





from math import exp, pow
from numpy import \
    array, ones, frompyfunc, \
    sqrt, linspace, where, trapz, \
    hstack
from traits.api import \
    Float,  \
    Instance, HasStrictTraits, on_trait_change,  \
    provides, Button, \
    Interface, WeakRef
from ibvpy.mathkit.mfn import MFnLineArray

class IPhiFn(Interface):

    '''Interface to adamage function representation
    '''

    def identify_parameters(self):
        '''Return the set of parameters defining the respective damage function.
        '''

    def __call__(self, e_max, *c_list):
        '''return the value of the damage function for the,
        maximum strain achieved so far and list of coefficients for material parameters
        '''

#-------------------------------------------------------------------------
# Damage function for MDM
#-------------------------------------------------------------------------


@provides(IPhiFn)
class PhiFnBase(HasStrictTraits):
    '''
    Damage function.
    '''
    polar_discr = WeakRef(transient=True)

    def __init__(self, **args):
        super(PhiFnBase, self).__init__(**args)
#        self.refresh_plot()

    mfn = Instance(MFnLineArray)

    def _mfn_default(self):
        return MFnLineArray(xdata=[0, 1], ydata=[1, 1])

    def refresh_plot(self):
        x_min, x_max = self.get_plot_range()
        x = linspace(x_min, x_max, 100)
        phi_fn = frompyfunc(self, 1, 1)
        x_ = x
        y_ = phi_fn(x)
        y = array([v for v in y_], dtype='float')
#        self.mfn.set(xdata = x, ydata = y)
#        self.mfn.data_changed = True

    def __call__(self, e_max, *c_list):
        return self.get_value(e_max, *c_list)

    def get_plot_range(self):
        '''
        Get the range of the plot domain
        '''
        raise NotImplementedError

    def identify_parameters(self):
        '''
        Extract the traits that are of type Float
        '''
        params = []
        for name, trait in list(self.traits().items()):
            if trait.is_trait_type(Float):
                params.append(name)
        return params

    def fit_params(self, *params):
        '''Possiblity to adapt the microplane-related
        material paramters based on the integral characteric specification.
        '''
        return

#-------------------------------------------------------------------------
# Piecewise Linear damage function for MDM
#-------------------------------------------------------------------------


class PhiFnGeneral(PhiFnBase):

    # implements(IPhiFn)

    def _polar_discr_changed(self):
        self.polar_discr.regularization = False

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        '''
        return self.mfn.get_value(e_max)

    def get_integ(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        '''
        # get the data that defines PhiFnGeneral
        # (fitted from experiment)
        _xdata = self.mfn.xdata
        _ydata = self.mfn.ydata

        # get the values smaller then the current e_max
        _xdata_ = _xdata[where(_xdata < e_max)[0]]
        _ydata_ = _ydata[:len(_xdata_)]
#        print '_xdata_' , _xdata_

        # add the value pair for e_max
        _xdata_emax = hstack([_xdata_, e_max])
        _ydata_emax = hstack([_ydata_, self.mfn.get_value(e_max)])
#        print '_xdata_emax' , _xdata_emax

        # assume an uncoupled relation (e.g. direct link) between the microplane
        # strains (e) and microplane stresses (s), e.g. s = phi * E * phi * e;
        # The methode 'get_integ' returns the value without the young's modulus,
        # which is multiplied in 'PhiFnPolar';
        # _ydata_integ = phi * phi * e
        #
        # @todo: this is only approximately true!; the correct evaluation
        # takes the version consistend (stiffness/compliance) pairs for
        # the microplane strains and stresses (work conjugates)
        _ydata_integ = _ydata_emax * _ydata_emax * _xdata_emax

        # integral under the stress-strain curve
        E_t = trapz(_ydata_integ, _xdata_emax)
        # area of the stored elastic energy
        U_t = 0.0
        if len(_xdata_emax) != 0:
            U_t = 0.5 * _ydata_integ[-1] * _xdata_emax[-1]
#        print 'E_t', E_t
#        print 'U_t', U_t
#        print 'E_t - U_t', E_t - U_t
        return E_t - U_t

    def get_plot_range(self):
        return self.mfn.xdata[0], self.mfn.xdata[-1]

    print_button = Button

    @on_trait_change('print_button')
    def print_button_fired(self):
        print('eps:\n', [self.mfn.xdata])
        print('1-omega:\n', [self.mfn.ydata])

#-------------------------------------------------------------------------
# Piecewise Linear damage function with drop to zero for MDM (used within 'MATSCalibDamageFn')
#-------------------------------------------------------------------------
class PhiFnGeneralExtended(PhiFnGeneral):

    # implements(IPhiFn)

    factor_eps_fail = Float(1.0)

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        Overload the 'get_value' method of 'PhiFnGeneral'
        and add an additional constant level and a drop
        down to zero after failure strain has been reached.
        '''
        eps_last = self.mfn.xdata[-1]
        phi_last = self.mfn.ydata[-1]
        eps_fail = eps_last * self.factor_eps_fail

        if e_max <= eps_last:
            return super(PhiFnGeneralExtended, self).get_value(e_max, *c_list)

        elif (e_max > eps_last and e_max < eps_fail):
            return phi_last
        else:
            return 1e-50

    def get_plot_range(self):
        '''plot the extended phi function'''
        return self.mfn.xdata[0], self.mfn.xdata[-1] * self.factor_eps_fail * 1.1


class PhiFnGeneralExtendedExp(PhiFnGeneral):

    # implements(IPhiFn)

    Dfp = Float(0.0, desc='residual integrity',
                enter_set=True, auto_set=False)
    Efp_frac = Float(0.01, desc='Efp factor',
                     enter_set=True, auto_set=False)

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        Overload the 'get_value' method of 'PhiFnGeneral'
        and add an exponential softening branch after
        failure strain has been reached.
        '''
        eps_last = self.mfn.xdata[-1]
        phi_last = self.mfn.ydata[-1]

        if e_max <= eps_last:
            return super(PhiFnGeneralExtendedExp, self).get_value(e_max)
        else:
            print('**** Entered softening branch ****')
            raise ValueError("Entered softening branch")
            # exponential softening with residual integrity after rupture
            # strain in the tensile test has been reached
            Dfp = self.Dfp
            Epp = eps_last
            Efp = self.Efp_frac * eps_last
            print('phi_last:', phi_last, ', eps_last:', eps_last, end=' ')
            phi = phi_last * \
                ((1 - Dfp) *
                 sqrt(Epp / e_max * exp(-(e_max - Epp) / Efp)) + Dfp)
            print(', e_max:', e_max, 'phi:', phi)
            return phi

    def get_plot_range(self):
        '''plot the extended phi function'''
        return self.mfn.xdata[0], self.mfn.xdata[-1] * 2.0

#-------------------------------------------------------------------------
# Damage function with stain softening for MDM
#-------------------------------------------------------------------------


class PhiFnStrainSoftening(PhiFnBase):

    '''
    Damage function.
    '''

    # implements(IPhiFn)

    G_f = Float(0.001117,
                label='G_f',
                desc='fracture energy',
                auto_set=False, enter_set=True)
    f_t = Float(2.8968,
                label='f_t',
                desc='tensile strength',
                auto_set=False, enter_set=True)
    md = Float(0.0,
               label='md',
               desc='factor affecting the compresive strength (explain more precisely)',
               auto_set=False, enter_set=True)
    h = Float(1.0,
              label='h',
              desc='element size to norm the fracture energy',
              auto_set=False, enter_set=True)

    Epp = Float(
        desc='strain at the onset of damage', enter_set=True, auto_set=False)
    Efp = Float(desc='strain at total damaged', enter_set=True, auto_set=False)

    @on_trait_change('G_f,f_t,md,h,polar_discr.E')
    def fit_microplane_params(self):
        '''
        Calculate the parameters of the damage function
        '''
        if self.polar_discr == None:
            return
        E = self.polar_discr.E
        G_f = self.G_f
        f_t = self.f_t
        md = self.md
        h = self.h

        gamma = (E * G_f) / (h * f_t ** 2)
        if gamma < 2.0:
            print('WARNING: elements too big -> refine, h should be at maximum only half of the characteristic length')
            print('in FIT PARAMS: gamma set to 2.0')
            gamma = 2.0

        Epp = f_t / \
            ((E * (1 - md) ** 2) * (1.95 - 0.95 / (gamma - 1) ** (0.5)))
        Efp = (G_f / ((1 - md) * h * E * Epp) +
               (2.13 - 1.13 * md) * Epp) / (2.73 - md) - Epp
        self.Epp = Epp
        self.Efp = Efp
        # @todo - plotting must be done separately
        # self.refresh_plot()

    def _polar_discr_changed(self):
        self.polar_discr.regularization = True

    def get_plot_range(self):
        return 0, self.Epp * 20.

    def get_integ(self, e_max, *c_list):
        '''
        OBSOLETE method - was used for decoupled evaluation of fracture
        energy contribution of the microplane.

        The method returns the value of the following integral:
        int( Phi(e_max~)**2 * e_max~, e_max~ = 0..e_max )
        The value corresponds to the fracture energy of the considered microplane
        divided by E. (Note: For a damage function Phi(e_max) the microplane stress-strain curve
        evaluates to s = Phi(e_max)*E*Phi(e_max)*e_max.)
        '''
        if len(c_list) == 0:
            c_list = [1., 1.]
        Epp = self.Epp * c_list[0]
        Efp = self.Efp * c_list[1]
        #
        # cf. derivation in Maple 'relation_between_Gf-ft_and_Epp-Efp
        # devide the result by E, i.e. the returned value does NOT include E
        if e_max <= Epp:
            return 0
        else:
            return -0.5 * Epp * (-Epp - 2.0 * Efp + 2.0 * Efp * exp(((-e_max + Epp) / Efp))) \
                   - 0.5 * e_max * Epp * exp(-((e_max - Epp) / Efp))

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        '''
        if len(c_list) == 0:
            c_list = [1., 1.]
        Epp = self.Epp * c_list[0]
        Efp = self.Efp * c_list[1]
        if e_max <= Epp:
            return 1.0
# @todo: check if this is necessary:
# if values smaller then 1.e-310 are returned
# a zero division error occures otherwise!
#        elif (e_max-Epp)/Efp >= 50:
#            return 1e-200
        else:
            return sqrt(Epp / e_max * exp(-(e_max - Epp) / Efp))


#-------------------------------------------------------------------------
# Damage function with residual damage level for MDM
#-------------------------------------------------------------------------
class PhiFnStrainHardeningLinear(PhiFnBase):

    '''
    Damage function which leads to a piecewise linear stress strain response.
    '''

    # implements(IPhiFn)

    E_f = Float(70e+3, desc='E-Modulus of the fibers',
                enter_set=True, auto_set=False, modified=True)
    E_m = Float(34e+3, desc='E-Modulus of the matrix',
                enter_set=True, auto_set=False, modified=True)
    rho = Float(0.03, desc='reinforcement ratio',
                enter_set=True, auto_set=False, modified=True)
    sigma_0 = Float(5.0, desc='asymptotic damage level',
                    enter_set=True, auto_set=False, modified=True)
    alpha = Float(0.0, desc='Slope of the strain hardening curve in section II',
                  enter_set=True, auto_set=False, modified=True)
    beta = Float(0.0, desc='Slope of the strain hardening curve in section III',
                 enter_set=True, auto_set=False, modified=True)
    Elimit = Float(0.006, desc='microplane strain at ultimate failure',
                   enter_set=True, auto_set=False, modified=True)

    def identify_parameters(self):
        return ['E_f', 'E_m', 'rho', 'sigma_0', 'alpha', 'beta', 'Elimit']

    def get_plot_range(self):
        return 0.0, self.Elimit * 1.2

    @on_trait_change('+modified')
    def refresh_plot(self):
        print('refreshing')
        super(PhiFnStrainHardeningLinear, self).refresh_plot()

    def _polar_discr_changed(self):
        print('regularizatoin set to False')
        self.polar_discr.regularization = False

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        '''
#        print 'x3c_list', c_list
        if len(c_list) == 0:
            c_list = [1., 1., 1., 1., 1., 1., 1.]
#        print 'x4c_list ', c_list

        E_f = self.E_f * c_list[0]
        E_m = self.E_m * c_list[1]
        rho = self.rho * c_list[2]
        sigma_0 = self.sigma_0 * c_list[3]
        alpha = self.alpha * c_list[4]
        beta = self.beta * c_list[5]
        Elimit = self.Elimit * c_list[6]
        #
        E_c = E_m * (1 - rho) + E_f * rho

        epsilon_0 = sigma_0 / E_c

        if e_max <= epsilon_0:
            return 1.0
        elif e_max >= Elimit:
            print('********** microplane reached maximum strain *********')
            return 1e-100

        epsilon_1 = sigma_0 * (-rho * E_f - E_m + rho * E_m + rho * E_f *
                               alpha + beta * E_m - beta * E_m * rho) / \
                              (rho * E_f + E_m - rho * E_m) / rho / E_f / \
                              (alpha - 1.0)

        epsilon = e_max

        if epsilon < epsilon_1:
            return sqrt(1.0 + (sigma_0 * rho * E_f + sigma_0 * E_m - sigma_0 * rho * E_m + rho * rho * E_f * E_f *
                               alpha * epsilon + rho * E_f * alpha * epsilon * E_m - rho * rho * E_f * alpha * epsilon * E_m - rho * E_f *
                               alpha * sigma_0 - epsilon * rho * rho * E_f * E_f - 2.0 * epsilon * rho * E_f * E_m + 2.0 * epsilon * rho *
                               rho * E_f * E_m - epsilon * E_m * E_m + 2.0 * epsilon * rho * E_m * E_m - epsilon * rho * rho * E_m * E_m) /
                        pow(rho * E_f + E_m - rho * E_m, 2.0) / epsilon)
        else:
            return sqrt(1.0 + E_m * (-E_f * rho * epsilon + epsilon * rho * rho * E_f + beta * sigma_0 - beta *
                                     sigma_0 * rho - epsilon * E_m + 2.0 * epsilon * rho * E_m - epsilon * rho * rho * E_m) /
                        pow(rho * E_f + E_m - rho * E_m, 2.0) / epsilon)

class PhiFnStrainHardening(PhiFnBase):

    '''
    Damage function.
    '''

    # implements(IPhiFn)

    Epp = Float(5.9e-05, desc='microplane strain at the onset of damage',
                enter_set=True, auto_set=False)
    Efp = Float(1.91e-04, desc='microplane strain at totaly damaged state',
                enter_set=True, auto_set=False)
    Dfp = Float(0.4, desc='asymptotic damage level',
                enter_set=True, auto_set=False)
    Elimit = Float(8.00e-02, desc='microplane strain at ultimate failure',
                   enter_set=True, auto_set=False)

    def identify_parameters(self):
        return ['Epp', 'Efp', 'Dfp', 'Elimit']

    def get_plot_range(self):
        return 0.0, self.Elimit * 1.2

    @on_trait_change('Epp,Efp,Dfp,Elimit')
    def refresh_plot(self):
        super(PhiFnStrainHardening, self).refresh_plot()

    def _polar_discr_changed(self):
        self.polar_discr.regularization = False

    def get_integ(self, e_max, *c_list):
        '''
        OBSOLETE method - was used for decoupled evaluation of fracture
        energy contribution of the microplane.

        The method returns the value of the following integral:
        int( Phi(e_max~)**2 * e_max~, e_max~ = 0..e_max )
        The value corresponds to the fracture energy of the considered microplane
        divided by E. (Note: For a damage function Phi(e_max) the microplane stress-strain curve
        evaluates to s = Phi(e_max)*E*Phi(e_max)*e_max.)
        '''
        if len(c_list) == 0:
            c_list = [1., 1.]
        Epp = self.Epp * c_list[0]
        Efp = self.Efp * c_list[1]
        Dfp = self.Dfp * c_list[2]
        Elimit = self.Elimit * c_list[3]
        # @todo: modify this for the case tension stiffening
        if e_max <= Epp:
            return 0
        else:
            return -0.5 * Epp * (-Epp - 2.0 * Efp + 2.0 * Efp * exp(((-e_max + Epp) / Efp))) \
                   - 0.5 * e_max * Epp * exp(-((e_max - Epp) / Efp))

    def get_value(self, e_max, *c_list):
        '''
        Evaluate the integrity of a particular microplane.
        '''
#        print 'x3c_list', c_list
        if len(c_list) == 0:
            c_list = [1., 1., 1., 1.]
#        print 'x4c_list ', c_list

#        print 'self.Epp used for TensionStiffening:', self.Epp
#        print 'self.Efp used for TensionStiffening:', self.Efp
#        print 'self.Dfp used for TensionStiffening:', self.Dfp
#        print 'self.Elimit used for TensionStiffening:', self.Elimit

        Epp = self.Epp * c_list[0]
        Efp = self.Efp * c_list[1]
        Dfp = self.Dfp * c_list[2]
        Elimit = self.Elimit * c_list[3]
        #
        if e_max <= Epp:
            return 1.0
        elif e_max >= Elimit:
            return 1.0e-100
# @todo: check if this is neccessary:
# if values smaller then 1.e-310 are returned
# a zero division error occures otherwise for Dfp=0!
#        elif (e_max-Epp)/Efp >= 50:
#            return Dfp
        else:
            return (1 - Dfp) * sqrt(Epp / e_max * exp(-(e_max - Epp) / Efp)) + Dfp


class PhiFnStrainHardeningBezier(PhiFnBase):

    '''Fitted polynomial'''

    epsilon_0 = Float(0.00004)
    epsilon_b = Float(0.0005)
    epsilon_f = Float(0.02)
    omega_b = Float(0.63)
    omega_f = Float(0.27)
    omega_t = Float(0.27)

    def identify_parameters(self):
        return ['epsilon_0', 'epsilon_b', 'epsilon_f', 'omega_b', 'omega_f', 'omega_t']

    def _polar_discr_changed(self):
        self.polar_discr.regularization = False

    def get_plot_range(self):
        return 0.0, self.epsilon_f * 1.5

    def get_value(self, epsilon, *c_list):

        if len(c_list) == 0:
            c_list = ones((6,), dtype=float)

        epsilon_0 = self.epsilon_0 * c_list[0]
        epsilon_b = self.epsilon_b * c_list[1]
        epsilon_f = self.epsilon_f * c_list[2]
        omega_b = self.omega_b * c_list[3]
        omega_f = self.omega_f * c_list[4]
        omega_t = self.omega_t * c_list[5]

        if epsilon <= epsilon_0:
            return 1.0
        elif epsilon_0 < epsilon and epsilon <= (epsilon_0 + epsilon_b):
            return 1 - omega_b / epsilon_b * (epsilon - epsilon_0)
        elif (epsilon_0 + epsilon_b) < epsilon and epsilon <= epsilon_0 + epsilon_b + epsilon_f:
            MapleGenVar1 = (pow(1.0 - (epsilon_b * omega_t - sqrt(epsilon_b * epsilon_b *
                                                                  omega_t * omega_t - 2.0 * epsilon_b * omega_t * epsilon * omega_b + 2.0 * epsilon_b * omega_t *
                                                                  epsilon_0 * omega_b + 2.0 * epsilon_b * epsilon_b * omega_t * omega_b + epsilon_f * omega_b *
                                                                  omega_b * epsilon - epsilon_f * omega_b * omega_b * epsilon_0 - epsilon_f * omega_b * omega_b *
                                                                  epsilon_b)) / (2.0 * epsilon_b * omega_t - epsilon_f * omega_b), 2.0) * omega_b)
            MapleGenVar3 = (2.0 * (1.0 - (epsilon_b * omega_t - sqrt(epsilon_b * epsilon_b *
                                                                     omega_t * omega_t - 2.0 * epsilon_b * omega_t * epsilon * omega_b + 2.0 * epsilon_b * omega_t *
                                                                     epsilon_0 * omega_b + 2.0 * epsilon_b * epsilon_b * omega_t * omega_b + epsilon_f * omega_b *
                                                                     omega_b * epsilon - epsilon_f * omega_b * omega_b * epsilon_0 - epsilon_f * omega_b * omega_b *
                                                                     epsilon_b)) / (2.0 * epsilon_b * omega_t - epsilon_f * omega_b)) * (epsilon_b * omega_t - sqrt(
                                                                         epsilon_b * epsilon_b * omega_t * omega_t - 2.0 * epsilon_b * omega_t * epsilon * omega_b + 2.0 *
                                                                         epsilon_b * omega_t * epsilon_0 * omega_b + 2.0 * epsilon_b * epsilon_b * omega_t * omega_b +
                                                                         epsilon_f * omega_b * omega_b * epsilon - epsilon_f *
                                                                         omega_b * omega_b *
                                                                         epsilon_0 -
                                                                         epsilon_f
                                                                         * omega_b * omega_b * epsilon_b)) / (2.0 * epsilon_b * omega_t - epsilon_f * omega_b) * (omega_b
                                                                                                                                                                  + omega_t))
            MapleGenVar4 = (pow(epsilon_b * omega_t - sqrt(epsilon_b * epsilon_b * omega_t *
                                                           omega_t - 2.0 * epsilon_b * omega_t * epsilon * omega_b + 2.0 * epsilon_b * omega_t * epsilon_0 *
                                                           omega_b + 2.0 * epsilon_b * epsilon_b * omega_t * omega_b + epsilon_f * omega_b * omega_b *
                                                           epsilon - epsilon_f * omega_b * omega_b * epsilon_0 -
                                                           epsilon_f * omega_b *
                                                           omega_b * epsilon_b
                                                           ), 2.0) / pow(2.0 * epsilon_b * omega_t - epsilon_f * omega_b, 2.0) * (omega_b + omega_f))
            MapleGenVar2 = MapleGenVar3 + MapleGenVar4
            return 1 - (MapleGenVar1 + MapleGenVar2)
        elif epsilon > epsilon_0 + epsilon_b + epsilon_f:
            eps_r = epsilon_0 + epsilon_b + epsilon_f
            omega = 1 - \
                (omega_b + omega_f + omega_b / epsilon_b * (epsilon - eps_r))
            if omega <= 0.001:
                return 0.001
            else:
                return omega



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch

from .mats3D_sdamage import \
    MATS3DScalarDamage

from ibvpy.tmodel.mats3D.__test__ import TestMATS3D

class TestMATS3DScalarDamage( TestMATS3D ):
    '''Test the elementary load cases of 3D material models.
    '''

    def _mats_eval_default(self):
        mats_eval = MATS3DScalarDamage( E = 34000, nu = 0.25 )   
        return mats_eval 
     
    def test_symmetry_tension(self):
        '''Symmetry of the response ... tensile loading.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [0,1,2], 
                                                        load = 0.001 )
                
    def test_symmetry_bending(self):
        '''Symmetry of the response ... out of plane load.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [1,2,0],
                                                        load = 0.000016 )
        return

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [ 1.6184728244667768, 0, 0, 0, 0, 0, 0, 0, 0 ],
                                    n_steps = 1, load = 0.0001 )
        
if __name__ == "__main__":
    import unittest
    import sys;sys.argv = ['', 'TestMATS3DScalarDamage.test_stress_value']         
    unittest.main()



'''
Created on 29.03.2017

@author: abaktheer

Microplane Fatigue model 3D

(compressive plasticity (CP) + tensile damage (TD) 
+ cumulative damage sliding (CSD))

Using Jirasek homogenization approach [1999]
'''

from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from numpy import array,\
    einsum, zeros_like, identity, sign,\
    sqrt
from traits.api import Constant,\
Property, cached_property, Dict, List
import bmcs_utils.api as bu
import numpy as np
import traits.api as tr


class MATS3DMplCSDEEQ(MATS3DEval):
    # implements(IMATSEval)

    # To use the model directly in the simulator specify the
    # time stepping classes

    node_name = 'Microplane CSD model (EEQ)'

    tree_node_list = List([])

    #---------------------------------------
    # Tangential constitutive law parameters
    #---------------------------------------
    gamma_T = bu.Float(10.,
                    label="Gamma",
                    desc=" Tangential Kinematic hardening modulus",
                    enter_set=True,
                    auto_set=False)

    K_T = bu.Float(0.0,
                label="K",
                desc="Tangential Isotropic harening",
                enter_set=True,
                auto_set=False)

    S_T = bu.Float(0.000001,
                label="S",
                desc="Damage strength",
                enter_set=True,
                auto_set=False)

    r_T = bu.Float(1.2,
                label="r",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    c_T = bu.Float(1.0,
                label="c",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    tau_pi_bar = bu.Float(3.0,
                       label="Tau_bar",
                       desc="Reversibility limit",
                       enter_set=True,
                       auto_set=False)

    a = bu.Float(0.0,
              label="a",
              desc="Lateral pressure coefficient",
              enter_set=True,
              auto_set=False)

    #-------------------------------------------
    # Normal_Tension constitutive law parameters (without cumulative normal strain)
    #-------------------------------------------
    Ad = bu.Float(100.0,
               label="Ad",
               desc="Brittleness coefficient",
               enter_set=True,
               auto_set=False)

    eps_0 = bu.Float(0.00002,
                  label="eps_0",
                  desc="Threshold strain",
                  enter_set=True,
                  auto_set=False)

    #-----------------------------------------------
    # Normal_Compression constitutive law parameters
    #-----------------------------------------------
    K_N = bu.Float(20.,
                label="K N compression",
                desc=" Normal isotropic harening",
                enter_set=True,
                auto_set=False)

    gamma_N = bu.Float(100000.,
                    label="gamma_compression",
                    desc="Normal kinematic hardening",
                    enter_set=True,
                    auto_set=False)

    sigma_0 = bu.Float(15.,
                    label="sigma 0 compression",
                    desc="Yield stress in compression",
                    enter_set=True,
                    auto_set=False)

    U_var_shape = (6,)
    '''Shape of the primary variable required by the TStepState.
    '''

    state_var_shapes = Property(Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''
    @cached_property
    def _get_state_var_shapes(self):
        return {'omegaN': (self.n_mp,),
                'z_N_Emn': (self.n_mp,),
                'alpha_N_Emn': (self.n_mp,),
                'r_N_Emn': (self.n_mp,),
                'eps_N_p_Emn': (self.n_mp,),
                'sigma_N_Emn': (self.n_mp,),
                'w_T_Emn': (self.n_mp,),
                'z_T_Emn': (self.n_mp,),
                'alpha_T_Emna': (self.n_mp, 3),
                'eps_T_pi_Emna': (self.n_mp, 3), }

    #--------------------------------------------------------------
    # microplane constitutive law (normal behavior CP + TD)
    #--------------------------------------------------------------
    def get_normal_law(self, eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn):

        E_N = self.E / (1.0 - 2.0 * self.nu)

        pos = eps_N_Emn > 1e-6
        H = 1.0 * pos

        sigma_n_trial = (1.0 - H * omegaN) * \
            E_N * (eps_N_Emn - eps_N_p_Emn)
        Z = self.K_N * r_N_Emn
        X = self.gamma_N * alpha_N_Emn

        h = self.sigma_0 + Z
        pos_iso = h > 1e-6
        f_trial = abs(sigma_n_trial - X) - h * pos_iso

        thres_1 = f_trial > 1e-6

        delta_lamda = f_trial / \
            (E_N + abs(self.K_N) + self.gamma_N) * thres_1
        eps_N_p_Emn = eps_N_p_Emn + delta_lamda * sign(sigma_n_trial - X)
        r_N_Emn = r_N_Emn + delta_lamda
        alpha_N_Emn = alpha_N_Emn + delta_lamda * sign(sigma_n_trial - X)

    # using the thermodynamic formulated tensile damage law
        def Z_N(z_N_Emn): return 1.0 / self.Ad * (-z_N_Emn) / (1.0 + z_N_Emn)
        Y_N = 0.5 * H * E_N * eps_N_Emn ** 2.0
        Y_0 = 0.5 * E_N * self.eps_0 ** 2.0
        f = Y_N - (Y_0 + Z_N(z_N_Emn))

        thres_2 = f > 1e-6

        thres_2 = f > 1e-6

#         def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))
#         omegaN = omegaN + f_w(Y_N) * thres_2
#         z_N_Emn = z_N_Emn - f_w(Y_N) * thres_2

#     # using Jirasek damage function
#         f_trial_Emn = eps_N_Emn - self.eps_0
#         f_idx = f_trial_Emn > 1e-6
#         f_idx = (f_idx * 1)


#     # using the thermodynamic formulated tensile damage law
#         def Z_N(z_N_Emn): return 1.0 / self.Ad * (-z_N_Emn) / (1.0 + z_N_Emn)
#         Y_N = 0.5 * H * E_N * eps_N_Emn ** 2.0
#         Y_0 = 0.5 * E_N * self.eps_0 ** 2.0
#         f = Y_N - (Y_0 + Z_N(z_N_Emn))

#
#         z_N_Emn = z_N_Emn.reshape(1, 28)
#         omegaN = omegaN.reshape(1, 28)
#
#         z_N_Emn[f_trial_Emn > 1e-6] = eps_N_Emn[f_trial_Emn > 1e-6]
#
#         omegaN[f_trial_Emn > 1e-6] = omegaN[f_trial_Emn > 1e-6] + (1. -
#                                                                    (self.eps_0 / z_N_Emn[f_trial_Emn > 1e-6]) * np.exp(- (z_N_Emn[f_trial_Emn > 1e-6] - self.eps_0) / (self.eps_f - self.eps_0)))
#

#         def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))
#         omegaN = omegaN + f_w(Y_N) * thres_2
#         z_N_Emn = z_N_Emn - f_w(Y_N) * thres_2

        thres_2 = f > 1e-6

    # using Jirasek damage function
        f_trial_Emn = eps_N_Emn - self.eps_0
        f_idx = np.where(f_trial_Emn > 0)
        # print(f_idx.shape)
        print(f_trial_Emn.shape)
        print(eps_N_Emn.shape)
        z_N_Emn[f_idx] = eps_N_Emn[f_idx]
        omegaN[f_idx] = (1. -
                         (self.eps_0 / z_N_Emn[f_idx]) * np.exp(- (z_N_Emn[f_idx] - self.eps_0) / (self.eps_f - self.eps_0)))

        def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))
        omegaN = f_w(Y_N) * thres_2
        z_N_Emn = - omegaN * thres_2


#     # using Jirasek damage function
#         f_trial_Emn = eps_N_Emn - self.eps_0
#         f_idx = np.where(f_trial_Emn > 0)
#         z_N_Emn[f_idx] = eps_N_Emn[f_idx]
#         omegaN[f_idx] = (1. -
#                          (self.eps_0 / z_N_Emn[f_idx]) * np.exp(- (z_N_Emn[f_idx] - self.eps_0) / (self.eps_f - self.eps_0)))
#

        sigma_N_Emn = (1.0 - H * omegaN) * E_N * (eps_N_Emn - eps_N_p_Emn)

        return omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn

    #-------------------------------------------------------------------------
    # microplane constitutive law (Tangential CSD)-(Pressure sensitive cumulative damage)
    #-------------------------------------------------------------------------
    def get_tangential_law(self, eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_kk):

        w_T_Emn = w_T_Emn.reshape(1, 28)
        E_T = self.E / (1.0 + self.nu)

        sig_pi_trial = E_T * (eps_T_Emna - eps_T_pi_Emna)
        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        norm_1 = sqrt(
            einsum('...na,...na->...n', (sig_pi_trial - X), (sig_pi_trial - X)))

        f = norm_1 - self.tau_pi_bar - \
            Z + self.a * sigma_kk

        plas_1 = f > 1e-6
        elas_1 = f < 1e-6

        delta_lamda = f / \
            (E_T / (1.0 - w_T_Emn) + self.gamma_T + self.K_T) * plas_1

        norm_2 = 1.0 * elas_1 + sqrt(
            einsum('...na,...na->...n', (sig_pi_trial - X), (sig_pi_trial - X))) * plas_1

        eps_T_pi_Emna[..., 0] = eps_T_pi_Emna[..., 0] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 0] - X[..., 0]) /
             (1.0 - w_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 1] = eps_T_pi_Emna[..., 1] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 1] - X[..., 1]) /
             (1.0 - w_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 2] = eps_T_pi_Emna[..., 2] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 2] - X[..., 2]) /
             (1.0 - w_T_Emn)) / norm_2

        Y = 0.5 * E_T * \
            einsum('...na,...na->...n', (eps_T_Emna - eps_T_pi_Emna),
                   (eps_T_Emna - eps_T_pi_Emna))

        w_T_Emn += ((1.0 - w_T_Emn) ** self.c_T) * \
            (delta_lamda * (Y / self.S_T) ** self.r_T)  # * \
        #(self.tau_pi_bar / (self.tau_pi_bar - self.a * sigma_kk / 3.0))

        alpha_T_Emna[..., 0] = alpha_T_Emna[..., 0] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 0] - X[..., 0]) / norm_2
        alpha_T_Emna[..., 1] = alpha_T_Emna[..., 1] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 1] - X[..., 1]) / norm_2
        alpha_T_Emna[..., 2] = alpha_T_Emna[..., 2] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 2] - X[..., 2]) / norm_2
        z_T_Emn = z_T_Emn + delta_lamda

        #eps_T_pi_Emna = np.zeros_like(eps_T_pi_Emna)

        return w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna

    #-------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    #-------------------------------------------------------------------------
    def _get_e_Emna(self, eps_Emab):
        # Projection of apparent strain onto the individual microplanes
        e_ni = einsum('nb,Emba->Emna', self._MPN, eps_Emab)
        return e_ni

    def _get_e_N_Emn(self, e_Emna):
        # get the normal strain array for each microplane
        e_N_Emn = einsum('Emna, na->Emn', e_Emna, self._MPN)
        return e_N_Emn

    def _get_e_T_Emna(self, e_Emna):
        # get the tangential strain vector array for each microplane
        e_N_Emn = self._get_e_N_Emn(e_Emna)
        e_N_Emna = einsum('Emn,na->Emna', e_N_Emn, self._MPN)
        return e_Emna - e_N_Emna

    #-------------------------------------------------
    # Alternative methods for the kinematic constraint
    #-------------------------------------------------
    # get the dyadic product of the microplane normals
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        MPNN_nij = einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')

    @cached_property
    def _get__MPTT(self):
        delta = identity(3)
        MPTT_nijr = 0.5 * (einsum('ni,jr -> nijr', self._MPN, delta) +
                           einsum('nj,ir -> njir', self._MPN, delta) - 2 *
                           einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN))
        return MPTT_nijr

    def _get_e_N_Emn_2(self, eps_Emab):
        # get the normal strain array for each microplane
        return einsum('nij,...ij->...n', self._MPNN, eps_Emab)

    def _get_e_T_Emna_2(self, eps_Emab):
        # get the tangential strain vector array for each microplane
        MPTT_ijr = self._get__MPTT()
        return einsum('nija,...ij->...na', MPTT_ijr, eps_Emab)

    def _get_e_Emna_2(self, eps_Emab):

        return self._get_e_N_Emn_2(eps_Emab) * self._MPN +\
            self._get_e_T_Emna_2(eps_Emab)

    #--------------------------------------------------------
    # return the state variables (Damage , inelastic strains)
    #--------------------------------------------------------
    def _get_state_variables(self, eps_Emab_n1, tn1,
                             omegaN, z_N_Emn,
                             alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
                             w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab_n1)
        eps_T_Emna = self._get_e_T_Emna_2(eps_Emab_n1)

        omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn = self.get_normal_law(
            eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
        w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        return omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna

    #-----------------------------------------------------------------
    # Returns a list of the plastic normal strain  for all microplanes.
    #-----------------------------------------------------------------
    def _get_eps_N_p_Emn(self, eps_Emab, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn):
        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_N_p_Emn = self.get_normal_law(
            eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[4]
        return eps_N_p_Emn

    #----------------------------------------------------------------
    # Returns a list of the sliding strain vector for all microplanes.
    #----------------------------------------------------------------
    def _get_eps_T_pi_Emna(self, eps_Emab, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_kk_Em):

        eps_T_Emna = self._get_e_T_Emna_2(eps_Emab)
        eps_N_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_kk_Em)[3]

        return eps_N_T_pi_Emna

    #-------------------------------------------------------------
    # Returns a list of the integrity factors for all microplanes.
    #-------------------------------------------------------------
    def _get_phi_Emn(self, eps_Emab, omegaN, z_N_Emn,
                     alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                     w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,  sigma_N_Emn):

        #         eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        #         eps_T_Emna = self._get_e_T_Emna_2(eps_Emab)
        #
        #         omegaN = self.get_normal_law(
        #             eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[0]
        #         sigma_N_Emn = self.get_normal_law(
        #             eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[5]
        #         w_T_Emn = self.get_tangential_law(
        # eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,
        # sigma_N_Emn)[0]

        w_Emn = zeros_like(omegaN)

#         w_Emn = np.zeros_like(omegaN)
#         idx_1 = np.where(sigma_N_Emn >= 0)
#         w_Emn[idx_1] = omegaN[idx_1]
#
#         idx_2 = np.where(sigma_N_Emn < 0)
#         w_Emn[idx_2] = w_T_Emn[idx_2]
#
        eig = np.linalg.eig(eps_Emab)[0]

        ter_1 = np.sum(eig)
        idx_1 = np.where(ter_1 > 0.0)
        idx_2 = np.where(ter_1 <= 0.0)

        #w_Emn = np.maximum(omegaN, w_T_Emn)
        w_Emn[idx_1] = omegaN[idx_1]
        w_Emn[idx_2] = w_T_Emn[idx_2]

        phi_Emn = sqrt(1.0 - w_Emn)

        print('omegaN', omegaN)
        print('w_T_Emn', w_T_Emn)
        print('w_Emn', w_Emn)
        print('phi_Emn', phi_Emn)

        return phi_Emn

    #----------------------------------------------
    # Returns the 2nd order damage tensor 'phi_mtx'
    #----------------------------------------------
    def _get_phi_Emab(self, phi_Emn):

        # integration terms for each microplanes
        phi_Emab = einsum('...n,n,nab->...ab', phi_Emn, self._MPW, self._MPNN)

        return phi_Emab

    #----------------------------------------------------------------------
    # Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
    #(cf. [Jir99], Eq.(21))
    #----------------------------------------------------------------------
    def _get_beta_Emabcd(self, phi_Emab):

        delta = identity(3)

        # use numpy functionality (einsum) to evaluate [Jir99], Eq.(21)
        beta_Emabcd = 0.25 * (einsum('...ik,jl->...ijkl', phi_Emab, delta) +
                              einsum('...il,jk->...ijkl', phi_Emab, delta) +
                              einsum('...jk,il->...ijkl', phi_Emab, delta) +
                              einsum('...jl,ik->...ijkl', phi_Emab, delta))

        return beta_Emabcd

    #---------------------------------------------------------------------
    # Extra homogenization of damage tensor in case of two damage parameters
    # Returns the 4th order damage tensor 'beta4' using (ref. [Baz99], Eq.(63))
    #---------------------------------------------------------------------

    def _get_beta_Emabcd_2(self, eps_Emab, omegaN, z_N_Emn,
                           alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                           w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,  sigma_N_Emn):

        # Returns the 4th order damage tensor 'beta4' using
        #(cf. [Baz99], Eq.(63))

        delta = identity(3)
        beta_N = sqrt(1. - omegaN)
        beta_T = sqrt(1. - w_T_Emn)

        beta_N = 1. - omegaN
        beta_T = 1. - w_T_Emn

        print('omegaN ', omegaN)
        print('omegaT ', w_T_Emn)
        print('beta_N ', beta_N)
        print('beta_T ', beta_T)

        beta_ijkl = einsum('n, ...n,ni, nj, nk, nl -> ...ijkl', self._MPW, beta_N, self._MPN, self._MPN, self._MPN, self._MPN) + \
            0.25 * (einsum('n, ...n,ni, nk, jl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    einsum('n, ...n,ni, nl, jk -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    einsum('n, ...n,nj, nk, il -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    einsum('n, ...n,nj, nl, ik -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) -
                    4.0 * einsum('n, ...n, ni, nj, nk, nl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, self._MPN, self._MPN))

        return beta_ijkl

    #-----------------------------------------------------------
    # Integration of the (inelastic) strains for each microplane
    #-----------------------------------------------------------
    def _get_eps_p_Emab(self, eps_Emab, omegaN, z_N_Emn,
                        alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                        w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,  sigma_N_Emn):

        delta = identity(3)

        # 2-nd order plastic (inelastic) tensor
        eps_p_Emab = einsum('n,...n,na,nb->...ab', self._MPW, eps_N_p_Emn, self._MPN, self._MPN) + \
            0.5 * (einsum('n,...nf,na,fb->...ab', self._MPW, eps_T_pi_Emna, self._MPN, delta) +
                   einsum('n,...nf,nb,fa->...ab', self._MPW, eps_T_pi_Emna, self._MPN, delta))

        #print ('eps_p_Emab', eps_p_Emab)

        return eps_p_Emab

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab_n1, tn1,
                      omegaN, z_N_Emn,
                      alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
                      w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):

        # Corrector predictor computation.
        # if update_state:

        #eps_Emab_n = eps_Emab_n1 - deps_Emab

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab_n1)
        eps_T_Emna = self._get_e_T_Emna_2(eps_Emab_n1)

        omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn = self.get_normal_law(
            eps_N_Emn, omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
        w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        #------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        #------------------------------------------------------------------
        phi_Emn = self._get_phi_Emn(eps_Emab_n1, omegaN, z_N_Emn,
                                    alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                    w_T_Emn, z_T_Emn, alpha_T_Emna,
                                    eps_T_pi_Emna, sigma_N_Emn)

        phi_Emab = self._get_phi_Emab(phi_Emn)

        #beta_Emabcd = self._get_beta_Emabcd(phi_Emab)

        beta_Emabcd = self._get_beta_Emabcd_2(eps_Emab_n1, omegaN, z_N_Emn,
                                              alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                              w_T_Emn, z_T_Emn, alpha_T_Emna,
                                              eps_T_pi_Emna, sigma_N_Emn)

        #------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        #------------------------------------------------------------------
        D_Emabcd = einsum(
            '...ijab, abef, ...cdef->...ijcd', beta_Emabcd, self.D_abef, beta_Emabcd)
        #----------------------------------------------------------------------
        # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
        #----------------------------------------------------------------------
        # plastic strain tensor
        eps_p_Emab = self._get_eps_p_Emab(eps_Emab_n1, omegaN, z_N_Emn,
                                          alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                          w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # elastic strain tensor
        eps_e_Emab = eps_Emab_n1 - eps_p_Emab

        # calculation of the stress tensor
        sig_Emab = einsum('...abcd,...cd->...ab', D_Emabcd, eps_e_Emab)
        # print(sig_Emab)

        return sig_Emab, D_Emabcd

    #-----------------------------------------------
    # number of microplanes - currently fixed for 3D
    #-----------------------------------------------
    n_mp = Constant(28)

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return array([[.577350259, .577350259, .577350259],
                      [.577350259, .577350259, -.577350259],
                      [.577350259, -.577350259, .577350259],
                      [.577350259, -.577350259, -.577350259],
                      [.935113132, .250562787, .250562787],
                      [.935113132, .250562787, -.250562787],
                      [.935113132, -.250562787, .250562787],
                      [.935113132, -.250562787, -.250562787],
                      [.250562787, .935113132, .250562787],
                      [.250562787, .935113132, -.250562787],
                      [.250562787, -.935113132, .250562787],
                      [.250562787, -.935113132, -.250562787],
                      [.250562787, .250562787, .935113132],
                      [.250562787, .250562787, -.935113132],
                      [.250562787, -.250562787, .935113132],
                      [.250562787, -.250562787, -.935113132],
                      [.186156720, .694746614, .694746614],
                      [.186156720, .694746614, -.694746614],
                      [.186156720, -.694746614, .694746614],
                      [.186156720, -.694746614, -.694746614],
                      [.694746614, .186156720, .694746614],
                      [.694746614, .186156720, -.694746614],
                      [.694746614, -.186156720, .694746614],
                      [.694746614, -.186156720, -.694746614],
                      [.694746614, .694746614, .186156720],
                      [.694746614, .694746614, -.186156720],
                      [.694746614, -.694746614, .186156720],
                      [.694746614, -.694746614, -.186156720]])

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return array([.0160714276, .0160714276, .0160714276, .0160714276, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505]) * 6.0

    #-------------------------------------------------------------------------
    # Cached elasticity tensors
    #-------------------------------------------------------------------------

    E = bu.Float(34e+3,
                 label="E",
                 desc="Young's Modulus",
                 auto_set=False,
                 input=True)

    nu = bu.Float(0.2,
                  label='nu',
                  desc="Poison ratio",
                  auto_set=False,
                  input=True)

    def _get_lame_params(self):
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_abef(self):
        la = self._get_lame_params()[0]
        mu = self._get_lame_params()[1]
        delta = identity(3)
        D_abef = (einsum(',ij,kl->ijkl', la, delta, delta) +
                  einsum(',ik,jl->ijkl', mu, delta, delta) +
                  einsum(',il,jk->ijkl', mu, delta, delta))

        return D_abef

    traits_view = bu.View(
                bu.Item('E'),
                bu.Item('nu'),
                bu.Item('gamma_T'),
                bu.Item('K_T'),
                bu.Item('S_T'),
                bu.Item('r_T'),
                bu.Item('c_T'),
                bu.Item('tau_pi_bar'),
                bu.Item('a'),
                bu.Item('Ad'),
                bu.Item('eps_0'),
                bu.Item('K_N'),
                bu.Item('gamma_N'),
                bu.Item('sigma_0'),
    )




from ibvpy.mesh.cell_grid.cell_spec import CellSpec
from ibvpy.mesh.cell_grid.cell_grid import CellGrid
from numpy import frompyfunc, c_, array, cos, sin
from math import pi

#def shell_stb( x, y ):
#    return x, y, x**2 - y**2
#    return x, y, 4e-5*x**2 - 0.2495*x - 1.4554 + -1e-6*y**2 + 0.0297*y -374.84
#
#vfn_shell_stb = frompyfunc( shell_stb, 2, 3 )
#
#def shell_stb_arr( points ):
#    x,y,z = vfn_shell_stb( points[:,0], points[:,1] )
#    points = c_[ array(x,dtype='float64'),
#                 array(y,dtype='float64'),
#                 array(z,dtype='float64') ]
#    return points
#
#mgd = CellGrid( shape   = (2,2),
#                coord_min = (-3,-10, 0),
#                coord_max = ( 3, 10, 0),
#                geo_transform  = shell_stb_arr,
#                grid_cell_spec = CellSpec( node_coords = [[-1,-1],[1,-1],[1,1],[-1,1]] ) )
#
from ibvpy.mesh.fe_grid import FEGrid


#def screwed_arch_3d( points ):
#    x = points[:,0]
#    y = points[:,1]
#    z = points[:,2]
#    l = x[-1] - x[0]
#    R = 10.
#    phi = x / l * pi
#    r = R + y
#    x,y,z = - r * cos( phi ), r * sin( phi ), z 
#    return c_[ x,y,z ]

def arch_3d( points ):
    x = points[:,0]
    y = points[:,1]
    z = points[:,2]
    D_z = 3.
    R = 30.
    alpha = pi/2.
    len_x = x[-1] - x[0]
    len_z = z[-1] - z[0]
    zz = z - len_z / 2.
    yy = y + 4 * D_z / len_z**2 * zz**2 + D_z
    
    phi_min, phi_max = pi/2. - alpha/2., pi/2. + alpha/2. 
    delta_phi = phi_max - phi_min 
    phi = phi_min + (x / len_x) * delta_phi
    r = R + yy
    #return c_[ x, yy, zz ]
    x,y,z = - r * cos( phi ), r * sin( phi ), zz * ( (phi - delta_phi) **2 + 2. ) 
    return c_[ x,y, z ]

#def arch_2d( points ):
#    x = points[:,0]
#    y = points[:,1]
#    l = x[-1] - x[0]
#    R = 10.
#    phi = x / l * pi
#    r = R + y
#    x,y = - r * cos( phi ), r * sin( phi ) 
#    return c_[ x,y ]
#
#fe_domain_arch_2d = FEGrid( geo_r = [[-1,-1 ],
#                                        [-1, 1 ],
#                                        [ 1,-1 ],
#                                        [ 1, 1 ]],                            
#                          dof_r = [[-1,-1 ],
#                                        [-1, 1 ],
#                                        [ 1,-1 ],
#                                        [ 1, 1 ]],
#                          shape = ( 10, 3 ),
#                          geo_transform = arch_2d )

if __name__ == '__main__':

    from ibvpy.fets.fets_eval import FETSEval
    fets_eval = FETSEval( geo_r = [[-1,-1,-1],
                                                [-1,-1, 1],
                                                [-1, 1,-1],
                                                [-1, 1, 1],
                                                [ 1,-1,-1],
                                                [ 1,-1, 1],
                                                [ 1, 1,-1],
                                                [ 1, 1, 1]],                            
                          dof_r = [[-1, 0, 0],
                                        [ 1, 0, 0],
                                        [ 0, 0,-1],
                                        [ 0, 0, 1],
                                        [ 0,-1, 0],
                                        [ 0, 1, 0]] )
    fe_domain_arch_3d = FEGrid( fets_eval = fets_eval,
                                      coord_max = ( 10., 1, 10. ),
                                      shape = ( 10, 3, 10 ),
                                      geo_transform = arch_3d )

    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = fe_domain_arch_3d )
    ibvpy_app.main()




from numpy import array, zeros, arange
from ibvpy.mathkit.linalg.dense_mtx import DenseMtx
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxAssembly


# bar clamped at the left end and loaded at the right end
def get_bar_mtx_array(shape):
    '''Get an array of matrices and dof_maps corresponding to
    a bar discretization
    '''
    el_mtx = array([[10, -10],
                    [-10, 10]], dtype='float64')

    el_mtx_arr = array([el_mtx for i in range(shape)], dtype=float)
    el_dof_map = array([arange(shape),
                        arange(shape) + 1], dtype=int).transpose()

    return el_dof_map, el_mtx_arr


def bar1():
    print('---------------------------------------------------------------')
    print('Clamped bar loaded at the right end with unit displacement')
    print('[00]-[01]-[02]-[03]-[04]-[05]-[06]-[07]-[08]-[09]-[10]')
    print('u[0] = 0, u[10] = 1')
    K = SysMtxAssembly()
    dof_map, mtx_arr = get_bar_mtx_array(shape=10)
    K.add_mtx_array(dof_map_arr=dof_map, mtx_arr=mtx_arr)
    K.register_constraint(a=0,  u_a=0.)  # clamped end
    K.register_constraint(a=10, u_a=1.)
    K.register_constraint(a=10, u_a=1.)
    K_dense = DenseMtx(assemb=K)
    R = zeros(K.n_dofs)
    print('K\n', K_dense)
    print('R\n', R)
    print('K_arrays')
    for i, sys_mtx_array in enumerate(K.sys_mtx_arrays):
        print('i\n', sys_mtx_array.mtx_arr)
    K.apply_constraints(R)
    K_dense = DenseMtx(assemb=K)
    print('K\n', K_dense)
    print('R\n', R)
    print('K_arrays')
    for i, sys_mtx_array in enumerate(K.sys_mtx_arrays):
        print('i\n', sys_mtx_array.mtx_arr)
    print('u =',  K.solve(R))
    print()


def bar2():
    print('---------------------------------------------------------------')
    print('Clamped bar composed of two linked bars loaded at the right end')
    print('[00]-[01]-[02]-[03]-[04]-[05]-[06]-[07]-[08]-[09]-[10]')
    print('[11]-[12]-[13]-[14]-[15]-[16]-[17]-[18]-[19]-[20]-[21]')
    print('u[0] = 0, u[5] = u[16], R[-1] = R[21] = 10')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=10)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    dof_map2, mtx_arr2 = get_bar_mtx_array(shape=10)
    # Note the dof map of the second br must be adjusted to start
    # the DOF enumeration with 3
    n_dofs1 = K.n_dofs
    K.add_mtx_array(dof_map_arr=dof_map2 + n_dofs1, mtx_arr=mtx_arr2)
    # add constraints
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=5, alpha=[1], ix_a=[16])
    # add load
    R = zeros(K.n_dofs)
    R[-1] = 10
    print('u =', K.solve(R))
    print()
    print('---------------------------------------------------------------')
    print('Clamped bar composed of two linked bars control displ at right')
    print('u[0] = 0, u[5] = u[16], u[21] = 1')
    # Remove the load and put a unit displacement at the right end
    # Note, the load is irrelevant in this case and will be rewritten
    #
    K.register_constraint(a=21, u_a=1)
    print('u =', K.solve(R))
    print()


def bar3():
    print('---------------------------------------------------------------')
    print('Clamped bar with recursive constraints (load at right end)')
    print('[0]-[1]-[2]-[3]')
    print('u[1] = 0.5 * u[2], u[2] = 1 * u[3], R[3] = 11')
    K = SysMtxAssembly()
    dof_map, mtx_arr = get_bar_mtx_array(shape=3)
    K.add_mtx_array(dof_map_arr=dof_map, mtx_arr=mtx_arr)
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=1, alpha=[0.5], ix_a=[2])
    K.register_constraint(a=2, alpha=[1.0], ix_a=[3])
    R = zeros(K.n_dofs)
    R[3] = 1
    K.apply_constraints(R)
    print('u =', K.solve())
    print()
    print('---------------------------------------------------------------')
    print('Clamped bar with recursive constraints (displ at right end)')
    print('u[1] = 0.5 * u[2], u[2] = 1.0 * u[3], u[3] = 1')
    K.register_constraint(a=3, u_a=1)
    print('u =', K.solve(R))
    print()


def bar4():
    print('---------------------------------------------------------------')
    print('Clamped bar 3 domains, each with 2 elems (displ at right end)')
    print('[0]-[1]-[2] [3]-[4]-[5] [6]-[7]-[8]')
    print('u[0] = 0, u[2] = u[3], u[5] = u[6], u[8] = 1')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=2)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    dof_map2, mtx_arr2 = get_bar_mtx_array(shape=2)
    K.add_mtx_array(dof_map_arr=dof_map2 + 3, mtx_arr=mtx_arr2)
    dof_map3, mtx_arr3 = get_bar_mtx_array(shape=2)
    K.add_mtx_array(dof_map_arr=dof_map3 + 6, mtx_arr=mtx_arr3)
    # add constraints
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=2, alpha=[1], ix_a=[3])
    K.register_constraint(a=5, alpha=[1], ix_a=[6])
    K.register_constraint(a=8, u_a=1.)  # loaded end
    # add load
    R = zeros(K.n_dofs)
    print('u =', K.solve(R))
    print()
    ####


def bar5():
    print('---------------------------------------------------------------')
    print('Clamped bar with 4 elements. Elements 2-4 are reinforced ')
    print('with another bar with 3 elements')
    print('[0]-[1]-[2]-[3]-[4]')
    print('    [5]-[6]-[7]')
    print('u[0] = 0, u[1] = u[5], u[3] = u[7], u[4] = 1')
    # assemble the matrix
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    dof_map2, mtx_arr2 = get_bar_mtx_array(shape=2)
    K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
    # add constraints
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=1, alpha=[1], ix_a=[5])
    K.register_constraint(a=3, alpha=[1], ix_a=[7])
    K.register_constraint(a=4, u_a=1.)  # loaded end
    # add load
    R = zeros(K.n_dofs)
    print('u =', K.solve(R))
    print()


def bar6():
    print('---------------------------------------------------------------')
    print('Clamped bar with 4 elements. Elements 2-4 are reinforced ')
    print('with another bar with 1 element linked proportianally')
    print('[0]-[1]-[2]-[3]-[4]')
    print('      [5]-[6]')
    print('u[0] = 0, u[1] = u[5], u[3] = u[7], u[4] = 1')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    dof_map2, mtx_arr2 = get_bar_mtx_array(shape=1)
    K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
    # add constraints
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=5, alpha=[0.5, 0.5], ix_a=[1, 2])
    K.register_constraint(a=6, alpha=[0.5, 0.5], ix_a=[2, 3])
    K.register_constraint(a=4, u_a=1.)  # loaded end
    # add load
    R = zeros(K.n_dofs)
    print('u =', K.solve(R))
    print()
    ####


def bar7():
    print('---------------------------------------------------------------')
    print('Two clamped beams link in parallel')
    print('and loaded by force at right end')
    print('[5]-[6]-[7]-[8]-[9]')
    print('[0]-[1]-[2]-[3]-[4]')
    print('u[5] = u[0], u[0] = 0, u[4] = 0.5 * u[9], R[4] = 1')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    dof_map2, mtx_arr2 = get_bar_mtx_array(shape=4)
    K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
    # add load
    R = zeros(K.n_dofs)
    # load at the coupled end nodes is doubled
    R[9] = 1
    R[4] = 1
    # add constraints
    K.register_constraint(a=5, alpha=[1], ix_a=[0])
    K.register_constraint(a=0, u_a=0.)  # clamped end
    K.register_constraint(a=4, alpha=[0.5], ix_a=[9])
    # add load
    R = zeros(K.n_dofs)
    # load at the coupled end nodes is doubled
    R[9] = 1
    R[4] = 1
    print('u =', K.solve(R))
    print()
    ####


def bar8():
    print('---------------------------------------------------------------')
    print('Single clamped element with two constraints')
    print('within a single element')
    print('[0]-[1]')
    print('u[0] = u[1], u[1] = 1')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=1)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    # add constraints
    K.register_constraint(a=0, alpha=[1], ix_a=[1])
    K.register_constraint(a=1, u_a=1.)  # clamped end
    # add load
    R = zeros(K.n_dofs)
    print('u =', K.solve(R))
    print()
    ####


def bar9():
    print('---------------------------------------------------------------')
    print('Single clamped element with two constraints')
    print('within a single element')
    print('[0]-[1]-[2]-[3]')
    print('u[0] = u[1], u[1] = 1')
    K = SysMtxAssembly()
    dof_map1, mtx_arr1 = get_bar_mtx_array(shape=3)
    K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
    # add constraints
    K.register_constraint(a=0)
    K.register_constraint(a=3, u_a=1.)  # clamped end
    # add load
    R = zeros(K.n_dofs)
    print('u =', K.solve(R))
    print()
    K.register_constraint(a=1, alpha=[1], ix_a=[2])
    print('u =', K.solve(R))
    print()
    ####

if __name__ == '__main__':
    bar1()
#    bar2()
#    bar3()
#    bar4()
#    bar5()
#    bar6()
#    bar7()
#    bar8()
#    bar9()




from traits.api import \
    Float, \
    Instance, Trait, on_trait_change, Event, \
    Dict, Property
from ibvpy.tmodel.mats1D.mats1D_eval import MATS1DEval
from ibvpy.mathkit.mfn import MFnLineArray
import numpy as np


#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS1DElastic(MATS1DEval):
    '''
    Elastic Model.
    '''

    E = Float(1.,  # 34e+3,
              label="E",
              desc="Young's Modulus",
              auto_set=False, enter_set=True)

    #-------------------------------------------------------------------------
    # Piece wise linear stress strain curve
    #-------------------------------------------------------------------------
    _stress_strain_curve = Instance(MFnLineArray)

    def __stress_strain_curve_default(self):
        return MFnLineArray(ydata=[0., self.E],
                            xdata=[0., 1.])

    @on_trait_change('E')
    def reset_stress_strain_curve(self):
        self._stress_strain_curve = MFnLineArray(ydata=[0., self.E],
                                                 xdata=[0., 1.])

    stress_strain_curve = Property

    def _get_stress_strain_curve(self):
        return self._stress_strain_curve

    def _set_stress_strain_curve(self, curve):
        self._stress_strain_curve = curve

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #-------------------------------------------------------------------------
    # Private initialization methods
    #-------------------------------------------------------------------------
    def __init__(self, **kwtraits):
        '''
        Subsidiary arrays required for the integration.
        they are set up only once prior to the computation
        '''
        super(MATS1DElastic, self).__init__(**kwtraits)

    def new_cntl_var(self):
        return np.zeros(1, np.float64)

    def new_resp_var(self):
        return np.zeros(1, np.float64)

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1, *args, **kw):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        eps_n1 = float(eps_app_eng)
#        E   = self.E
#        D_el = array([[E]])
#        sigma = dot( D_el, eps_app_eng )
        D_el = np.array([[self.stress_strain_curve.diff(eps_n1)]])
        sigma = np.array([self.stress_strain_curve(eps_n1)])
        # You print the stress you just computed and the value of the apparent
        # E
        return sigma, D_el

    #-------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #-------------------------------------------------------------------------

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app}




'''
Created on Dec 17, 2018

'''
from traits.api import Interface


class ITLoop(Interface):
    '''Loop performing the simulation of a time-stepping model  
    along the specified time line and recording its history.
    '''

    def __call__(self):
        '''Implementation of the loop.
        '''



'''
Created on Sep 3, 2009

@author: jakub
'''

from math import pi, sqrt
from scipy.linalg import \
    inv
from traits.api import \
    Instance, Property, \
    DelegatesTo

from ibvpy.fets.fets_eval import \
    FETSEval
import numpy as np


#-------------------------------------------------------------------------
# FETS2D4Q - 4 nodes iso-parametric quadrilateral element (2D, linear, Lagrange family)
#-------------------------------------------------------------------------
# class FETS2Drotsym( FETSEvalPrototyped ):
class FETS2Drotsym(FETSEval):

    debug_on = True

    prototype_fets = Instance(FETSEval)

    # Dimensional mapping
    dim_slice = slice(0, 2)
    dof_r = DelegatesTo('prototype_fets')
    geo_r = DelegatesTo('prototype_fets')
    n_nodal_dofs = DelegatesTo('prototype_fets')
    n_e_dofs = DelegatesTo('prototype_fets')

    get_dNr_mtx = DelegatesTo('prototype_fets')
    get_dNr_geo_mtx = DelegatesTo('prototype_fets')

    get_N_geo_mtx = DelegatesTo('prototype_fets')
    get_N_mtx = DelegatesTo('prototype_fets')

    ngp_r = Property

    def _get_ngp_r(self):
        return self.prototype_fets.ngp_r

    ngp_s = Property

    def _get_ngp_s(self):
        return self.prototype_fets.ngp_s

    vtk_cell_types = Property

    def _get_vtk_cell_types(self):
        return self.prototype_fets.vtk_cell_types

    vtk_cells = Property

    def _get_vtk_cells(self):
        return self.prototype_fets.vtk_cells

    vtk_r = Property

    def _get_vtk_r(self):
        return self.prototype_fets.vtk_r

    def _set_vtk_r(self, value):
        # @todo - WHAT's this!!!! Setter should put the value in and not return anything!
        return value

    def get_B_mtx(self, r_pnt, X_mtx):
        '''Mapping matrix from displacements to strains
        '''
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        N_mtx = self.get_N_mtx(r_pnt)
        radius = np.dot(self.get_N_geo_mtx(r_pnt), X_mtx[:, 1])

        dNx_mtx = np.dot(inv(J_mtx), dNr_mtx)

        Bx_mtx = np.zeros((6, self.n_e_dofs), dtype='float64')
        for i in range(0, self. n_dof_r):
            Bx_mtx[0, i * 2] = dNx_mtx[0, i]  # eps_z
            Bx_mtx[1, i * 2 + 1] = dNx_mtx[1, i]  # eps_r
            Bx_mtx[2, i * 2 + 1] = N_mtx[1, i * 2 + 1] / radius  # eps_theta
            Bx_mtx[5, i * 2] = dNx_mtx[1, i]  # gamma_rz
            Bx_mtx[5, i * 2 + 1] = dNx_mtx[0, i]  # gamma_rz
        return Bx_mtx

    def get_J_det(self, r_pnt, X_mtx):
        '''
        Has to be overloaded for rotational symetry case
        @param r_pnt:
        @param X_mtx:
        '''
        #radius = 1.
        circle = 2. * pi * np.dot(self.get_N_geo_mtx(r_pnt), X_mtx[:, 1])
        return np.array(circle * self._get_J_det(r_pnt, X_mtx), dtype='float64')

    def adjust_spatial_context_for_point(self, sctx):
        '''Overloaded call to mats_eval.

        X_mtx is extended, the depth component is the circle node 
        draws around the axis of symmetry

        The extension is required in order to regularize the process zone
        within the element.
        '''
        if sctx.X.shape[1] == 2:
            X_mtx_theta = sctx.X[:, 1] * 2 * pi
            sctx.X_reg = np.append(
                np.copy(sctx.X), X_mtx_theta[:, None], axis=1)

#----------------------- example --------------------


def example_with_new_domain():
    from ibvpy.api import \
        TStepper as TS, RTraceDomainListField, TLoop, \
        TLine, BCSlice
    from ibvpy.fets.fets2D.fets2D4q8u import FETS2D4Q8U
    from ibvpy.tmodel.mats3D.mats3D_cmdm import \
        MATS3DMicroplaneDamage
    from ibvpy.tmodel.matsXD.matsXD_cmdm import PhiFnStrainSoftening

#    tmodel =  MATS2DElastic(E=2,nu= .2,
#                          stress_state= 'rotational_symetry')
    mats = MATS3DMicroplaneDamage(model_version='stiffness',
                                  E=34e3,
                                  nu=0.2,
                                  phi_fn=PhiFnStrainSoftening(G_f=0.001117,
                                                              f_t=2.8968))

    fets_eval = FETS2Drotsym(prototype_fets=FETS2D4Q8U(),
                             mats_eval=mats)

    fets_eval.vtk_r *= 0.9
    from ibvpy.mesh.fe_grid import FEGrid

    radius = sqrt(1. / pi)
#    f_i = (radius/2.)*2*pi
#    f_o = (radius)*2*pi
#    print 'f ',f_i,' ', f_o
    # Discretization
    fe_grid = FEGrid(  # coord_min = (0.,radius/2.,0.),
        coord_max=(1., radius, 0.),
        shape=(20, 20),
        fets_eval=fets_eval)

    tstepper = TS(sdomain=fe_grid,
                  bcond_list=[
                      BCSlice(var='u', value=0., dims=[0],
                              slice=fe_grid[0, :, 0, :]),
                      BCSlice(var='u', value=0., dims=[1],
                              slice=fe_grid[0, 0, 0, 0]),
                      BCSlice(var='u', value=1.e-3, dims=[0],
                              slice=fe_grid[-1, :, -1, :]),
                  ],

                  rtrace_list=[
                      RTraceDomainListField(name='Stress',
                                            var='sig_app', idx=0, warp=True,
                                            record_on='update'),
                      RTraceDomainListField(name='fracture_energy',
                                            var='fracture_energy', idx=0, warp=True,
                                            record_on='update'),
                      RTraceDomainListField(name='Displacement',
                                            var='u', idx=0,
                                            record_on='update',
                                            warp=True),
                      #                    RTraceDomainListField(name = 'N0' ,
                      #                                      var = 'N_mtx', idx = 0,
                      # record_on = 'update')
                  ]
                  )

    # Add the time-loop control
    #global tloop
    tloop = TLoop(tstepper=tstepper, KMAX=300, tolerance=1e-4,
                  tline=TLine(min=0.0,  step=1.0, max=1.0))

    #import cProfile
    #cProfile.run('tloop.eval()', 'tloop_prof' )
    print(tloop.eval())
    #import pstats
    #p = pstats.Stats('tloop_prof')
    # p.strip_dirs()
    # print 'cumulative'
    # p.sort_stats('cumulative').print_stats(20)
    # print 'time'
    # p.sort_stats('time').print_stats(20)

    # Put the whole thing into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tloop)
    app.main()

if __name__ == '__main__':
    example_with_new_domain()





from traits.api import HasTraits, Array, Property, cached_property
from numpy import allclose, arange, eye, linalg, ones, ix_, array, zeros, \
                hstack,meshgrid, vstack, dot, newaxis, c_, r_, copy, where,\
                ones
from scipy import sparse,linalg
from scipy.sparse.linalg.dsolve import linsolve
from time import time

class SysDenseMtx( HasTraits ):
    """
    Block of element matrices
    """
    def __init__(self, n_dofs, el_dof_map ):
        self.el_dof_map = el_dof_map
        self.mtx = zeros( [n_dofs, n_dofs], dtype = float )

    def set_data(self, data_l ):
        for el_dof_map, el_mtx in zip( self.el_dof_map, data_l ):
            self.mtx[ ix_( el_dof_map, el_dof_map ) ] += el_mtx
        
    def set_nond_zero(self, dof):
        K_aa = self.mtx[dof,dof]
        self.mtx[:,dof] = 0.
        self.mtx[dof,:] = 0.
        self.mtx[dof, dof]   = - K_aa
    
    def add_column(self, dof, F, factor):   
        F += factor * self.mtx[:, dof]
    
    def get_diag_elem(self, dof):
        return self.mtx[dof,dof]
    
    def adapt_linked_mtx(self, dof, n, alpha, K_aa):
        Kna = self._const_dof_diag(dof, alpha)
        self.set_nond_zero(dof)
        self._adapt_linked_stiff(dof, n, alpha, Kna, K_aa)
    
    def _const_dof_diag(self, dof, alpha):
        return dot( self.mtx[:,[dof]], alpha[newaxis] )
    
    def _adapt_linked_stiff(self, dof, n, alpha, Kna, K_aa ):
        a_n_ix = ix_( [dof], n )  # array index for submatrix [a,n]
        n_a_ix = ix_( n, [dof] )  # array index for submatrix [n,a]
        
        self.mtx[:,n] += Kna
        self.mtx[n,:] += Kna.transpose()

        # Build in the kinematic constraint 
        #
        #print "comparison ",self.mtx[n_a_ix]," ",alpha.transpose() * K_aa
        self.mtx[n_a_ix] = alpha[newaxis] .transpose() * K_aa
        self.mtx[a_n_ix] = alpha[newaxis] * K_aa

    def solve(self, rhs):
        u_vct = linalg.solve( self.mtx, rhs )
        return u_vct
    
class SysSparseMtx( HasTraits ):
    
    el_dof_map = Array( int )
    
    def __init__(self, n_dofs, el_dof_map ):
        '''Construct the symmetric sparse matrix with el_dof_map
        defining the sparsity map. The assembly is done from a list
        of element matrices that follow the same order 
        as the one in el_dof_map.''' 
        self.n_dofs = n_dofs
        self.el_dof_map = el_dof_map
        self.Kna_data = []

    def set_data(self, mtx_arr ):
        '''Set the current array of matrices to be assembled in 
        the sparce matrices. The order of the matrices follows that in 
        el_dof_map'''
        self.mtx_arr = mtx_arr

    ij_map = Property( depends_on = 'el_dof_map' )
    @cached_property
    def _get_ij_map( self ):
        '''
        Derive the row and column indices of individual values 
        in every element matrix.
        '''
        el_dof_map = self.el_dof_map
        ij_dof_map = zeros( (el_dof_map.shape[0],
                             2,
                             el_dof_map.shape[1]**2,
                             ), dtype = 'int_' )
        for el, dof_map in enumerate( el_dof_map ):
            row_dof_map, col_dof_map = meshgrid(dof_map,dof_map)
            ij_dof_map[el,...] = vstack( [row_dof_map.flatten(), col_dof_map.flatten()] )
        return ij_dof_map 

    x_l = Property( depends_on = 'el_dof_map' )
    @cached_property
    def _get_x_l(self):
        '''Helper property to get an array of all row indices'''
        return self.ij_map[:,0,:].flatten()
    
    y_l = Property( depends_on = 'el_dof_map' )
    @cached_property
    def _get_y_l(self):
        '''Helper property to get an array of all column indices'''
        return self.ij_map[:,1,:].flatten()
                        
    def set_nond_zero(self, dof):
        '''Set the off-diagonal values associated with the row and column a
        to zero. Remark, set the diagonal terms negative. 
        
        @TODO this can be improved using the where method of numpy
        '''

        el_arr, row_arr = where( self.el_dof_map == dof )
        for el, i_dof in zip( el_arr, row_arr ):        
            k_diag = self.mtx_arr[el,i_dof,i_dof]
            self.mtx_arr[el,i_dof,:] = 0.0
            self.mtx_arr[el,:,i_dof] = 0.0
            self.mtx_arr[el,i_dof,i_dof] = -k_diag
                
    def add_column(self, dof, F, factor):   
        '''Get the slice of the a-th column.
        (used for the implementation of the essential boundary conditions)
        @TODO use where method of numpy 
        '''             
        el_arr, row_arr = where( self.el_dof_map == dof )
        for el, i_dof in zip( el_arr, row_arr ):
            rows = self.el_dof_map[el]
            F[ rows ] += factor * self.mtx_arr[ el, :, i_dof ]
        
    def get_diag_elem(self, dof):
        '''Get the value of diagonal element at a-ths dof. 
        '''
        K_aa = 0.
        el_arr, row_arr = where( self.el_dof_map == dof )
        for el, i_dof in zip( el_arr, row_arr ):
            K_aa += self.mtx_arr[el, i_dof, i_dof ]
        return K_aa
    
    def adapt_linked_mtx(self, dof, n, alpha, K_aa):
        '''Get the submatrix of the constrained degrees of freedom.
        '''
        Kna_idx = []
        self.Kna_data = []
        el_a_arr, row_a_arr = where( self.el_dof_map == dof )
        for el, i_dof in zip( el_a_arr, row_a_arr ):
            rows = self.el_dof_map[el]
            sl = self.mtx_arr[ el, :, [i_dof] ]
            #Kna[ rows ] += dot(sl.T,alpha)
            self.Kna_data.extend(dot(sl.T,alpha))
            Kna_idx.extend(rows)
            k_diag = self.mtx_arr[el,i_dof,i_dof]
            self.mtx_arr[el,i_dof,:] = 0.0
            self.mtx_arr[el,:,i_dof] = 0.0
            self.mtx_arr[el,i_dof,i_dof] = -k_diag
        
        Kna_n_idx = ones(len(Kna_idx)) * n
        self.Kna_ij = vstack((hstack((Kna_n_idx,Kna_idx)),\
                              hstack((Kna_idx,Kna_n_idx))))
        
        self.Kna_data = hstack((self.Kna_data,self.Kna_data))
            
    def solve(self, rhs):
        '''Construct the matrix and use the solver to get 
        the solution for the supplied rhs. 
        '''
        self.data_l = self.mtx_arr.ravel()        
        if self.Kna_data != []:
            ij = hstack((vstack((self.x_l,self.y_l)),self.Kna_ij))
            self.data_l = hstack((self.data_l,self.Kna_data))
        else:
            ij = vstack((self.x_l,self.y_l))

        # Assemble the system matrix from the flattened data and 
        # sparsity map containing two rows - first one are the row
        # indices and second one are the column indices.
        mtx = sparse.coo_matrix( ( self.data_l, ij ) )

        u_vct = linsolve.spsolve( mtx, rhs )
        return u_vct  

        
if __name__ == '__main__':

    # todo - run the solver for both bypes of matrices
    # and compare the solution to get a unittest.
    # 
    n_dofs = 10
    submtx = array([[1.,-1.],[-1.,1.]])
    
    el_dof_map = array( [arange( n_dofs - 1),
                         arange( n_dofs - 1) + 1 ], dtype = int ).transpose()

    rhs = zeros(n_dofs)
    rhs[-1]= 1.
    
#    tf_start = time()
#    mtx = SysDenseMtx( n_dofs )
#    for ix in ix_list:    
#        mtx[ ix_(ix,ix)  ] += submtx
#    
#   
#
#    mtx[0,:] = 0.
#    mtx[:,0] = 0.
#    mtx[0,0] = 1.
#    #print "mtx ",mtx
#    U_f=mtx.solve(rhs)
#    print "U_f ",U_f
#    tf_end = time()
#    diff = tf_end - tf_start
#    print "Full Matrix: %8.2f sec" %diff
    
    ts_start = time()

    
    x_l = []
    y_l = []
    data_l = []
    for ix in range(n_dofs-1):
        data_l.append(submtx)
    
    #print "X,Y ", x_a,y_a
    #print "data ", data_a
    #print "X,Y ",x_a,y_a
    #x_a = array([x_l]).flatten()
    #y_a = array([y_l]).flatten()
    #ix_arr = vstack((x_a,y_a))
    #print "ix_arr ", ix_arr
    #data = array(data_l).flatten()
    #print data.shape[0]
    sp_mtx = SysSparseMtx(n_dofs, el_dof_map )
     
    print('ij_map',sp_mtx.ij_map)
    
    sp_mtx.set_data( array( data_l ) )
    #sp_mtx.set_nond_row_zero(0)
    #sp_mtx.set_nond_col_zero(0)
    sp_mtx.set_nond_zero(0)
    print("class ", sp_mtx.__class__)
    #K_aa =  sp_mtx[0,:]       
    #sp_mtx[0] = 0.
    #sp_mtx[:,0] = 0.
    #sp_mtx[0,0] = 1.
    print("K_aa ",sp_mtx.get_diag_elem(2))
    U_s = sp_mtx.solve(rhs)
    print("U_s ", U_s)
    ts_end = time()
    difs = ts_end - ts_start
    print("Sparse Matrix: %8.2f sec" %difs)

#    for i in range(5,0,-1):
#        print "i ",i



from .mats_eval import MATSEval, IMATSEval
from .mats1D import \
    MATS1DElastic, MATS1DPlastic, MATS1DDamage
from .mats1D5 import MATS1D5BondSlipMultiLinear, MATS1D5D, \
    MATS1D5BondSlipD, MATS1D5BondSlipEP, MATS1D5BondSlipTriLinear
from .mats2D.mats2D_cmdm.mats2D_cmdm import MATS2DMicroplaneDamage
from .mats3D.mats3D_cmdm.mats3D_cmdm import MATS3DMicroplaneDamage
from .matsXD.matsXD_cmdm.matsXD_cmdm_phi_fn import \
    PhiFnGeneral, PhiFnGeneralExtended, PhiFnStrainSoftening, \
    PhiFnStrainHardening, PhiFnStrainHardeningBezier, PhiFnStrainHardeningLinear



from numpy import zeros, hstack, meshgrid, vstack
from scipy import sparse
from scipy.sparse.linalg import eigsh, spsolve
from traits.api import HasTraits, Property, cached_property, Any

import numpy as np


class COOSparseMtx(HasTraits):

    assemb = Any

    ij_map = Property(depends_on='assemb.+')
    @cached_property
    def _get_ij_map(self):
        '''
        Derive the row and column indices of individual values
        in every element matrix.
        '''

        O_list, P_list = [], []
        # loop over the list of matrix arrays
        for sys_mtx_arr in self.assemb.get_sys_mtx_arrays():

            O_Eo = sys_mtx_arr.dof_map_arr
            a = np.arange(2, dtype=np.int_)
            O, P = (np.einsum('a,Lo->aLo', (1 - a), O_Eo)[:, :, :, None] +
                    np.einsum('a,Lp->aLp', a, O_Eo)[:, :, None, :])
            O_list.append(O.flatten())
            P_list.append(P.flatten())

        return np.hstack(O_list), np.hstack(P_list)

    data_l = Property

    def _get_data_l(self):

        return hstack([sm_arr.mtx_arr.ravel()
                       for sm_arr in self.assemb.get_sys_mtx_arrays()])

    def solve(self, rhs, check_pos_dev=False):
        '''Construct the matrix and use the solver to get 
        the solution for the supplied rhs
        pos_dev - test the positive definiteness of the matrix. 
        '''

        # Assemble the system matrix from the flattened data and
        # sparsity map containing two rows - first one are the row
        # indices and second one are the column indices.
        mtx = sparse.coo_array((self.data_l, self.ij_map))
        mtx_csr = mtx.tocsr()

        pos_def = True
        if check_pos_dev:
            evals_small, evecs_small = eigsh(mtx_csr, 3, sigma=0, which='LM')
            min_eval = np.min(evals_small)
            pos_def = min_eval > 1e-10

        u_vct = spsolve(mtx_csr, rhs)
        return u_vct, pos_def

    def toarray(self):
        return sparse.coo_array((self.data_l, self.ij_map)).toarray()
